#!/usr/bin/env python3
"""
TODO:
    - [ ] Rename to stac_to_kwcoco

SeeAlso:
    ~/code/watch/watch/cli/stac_search.py
"""
import os
import re
import sys
import ubelt as ub
import scriptconfig as scfg

from os.path import basename, dirname, join

try:
    from xdev import profile
except Exception:
    profile = ub.probile


class StacToCocoConfig(scfg.DataConfig):
    """
    Convert a STAC catalog to a KWCOCO manifest
    """
    input_stac_catalog = scfg.Value(None, type=str, position=1, required=True, help='Input STAC Catalog to convert to KWCOCO')

    outpath = scfg.Value(None, type=str, short_alias=['o'], help='Output path for updated STAC catalog')

    assume_relative = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Assume the data is in subdirectories relative to the output
            KWCOCO manifest
            '''))

    from_collated = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Data to convert has been run through TA-1 collation
            '''))

    ignore_duplicates = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Ignore duplicate items when creating the kwcoco file
            '''))

    populate_watch_fields = scfg.Value(False, isflag=True, help='Populate video / watch fields')

    verbose = scfg.Value(1, help='verbosity')

    jobs = scfg.Value(1, type=str, short_alias=['j'], help='Number of jobs to run in parallel')


def main(cmdline=True, **kwargs):
    config = StacToCocoConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    ta1_stac_to_kwcoco(**config)
    return 0


SUPPORTED_COARSE_PLATFORMS = {
    'S2': {'S2A', 'S2B', 'sentinel-2a', 'sentinel-2b', 'S2'},  # Sentinel-2
    'L8': {'OLI_TIRS', 'LANDSAT_8', 'L8'},  # Landsat-8
    'WV': {'DigitalGlobe', 'worldview-2', 'worldview-3', 'WV', 'WV02', 'WV03'},  # Worldview
    'WV1': {'WV01'},  # Worldview 1
    'PD': {'PlanetScope', 'dove', 'PD'},  # Planet
}


def normalize_str(s):
    return s.lower().replace('-', '_')


PLATFORM_NORMALIZED_TO_PLATFORM_STANDARD_CASE = {
    normalize_str(v): v
    for vs in SUPPORTED_COARSE_PLATFORMS.values()
    for v in vs
}

SUPPORTED_PLATFORMS = set.union(
    *SUPPORTED_COARSE_PLATFORMS.values(),
    set(SUPPORTED_COARSE_PLATFORMS.keys()))


SENSOR_COARSE_MAPPING = {
    v: k
    for k, vals in SUPPORTED_COARSE_PLATFORMS.items()
    for v in vals
}


def _construct_sensor_channel_alias():
    """
    Construct mappings from possible names for each bands to the ones that we
    want to use in kwcoco.
    """
    from watch.utils import util_bands
    UTIL_BAND_INFOS = {
        'S2': util_bands.SENTINEL2,
        'L8': util_bands.LANDSAT8,
    }
    SENSOR_CHANNEL_ALIAS = {}
    for sensor_code, band_infos in UTIL_BAND_INFOS.items():
        alias_lut = SENSOR_CHANNEL_ALIAS[sensor_code] = {}
        for band in band_infos:
            if 'common_name' in band:
                alias_lut[band['name']] = band['common_name']
                alias_lut[band['common_name']] = band['common_name']
                for alias in band.get('alias', []):
                    alias_lut[alias] = band['common_name']
        # TCI is not a true band, but often included anyway
        # and this channel code is more specific to kwcoco
        alias_lut['TCI'] = 'tci:3'
    return SENSOR_CHANNEL_ALIAS


SENSOR_CHANNEL_ALIAS = _construct_sensor_channel_alias()


def _determine_channels_collated(asset_name, asset_dict, platform):
    """
    Note:
        The term "collated" means that each band is its own asset and it has
        the eo:bands property. For more details see:
        https://smartgitlab.com/TE/standards/-/wikis/STAC-and-Storage-Specifications
    """
    sensor_coarse = SENSOR_COARSE_MAPPING.get(platform, platform)

    eo_band_names = [eob.get('common_name', eob['name'])
                     for eob in asset_dict.get('eo:bands', ())]

    # Map aliases for this sensor if we have registered it
    channel_alias_lut = SENSOR_CHANNEL_ALIAS.get(sensor_coarse, None)
    if channel_alias_lut is not None:
        eo_band_names = [
            channel_alias_lut.get(name, name) for name in eo_band_names]

    if len(eo_band_names) > 0:
        return '|'.join(eo_band_names)
    elif asset_name == 'quality':
        return 'quality'
        # return 'cloudmask'


def _determine_s2_channels(asset_name, asset_dict):
    """
        >>> from watch.cli.ta1_stac_to_kwcoco import *  # NOQA
        >>> from watch.cli.ta1_stac_to_kwcoco import _determine_s2_channels
        >>> test_hrefs = [
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/15/T/TF/2020/9/21/S2A_14TQL_20200921_0_L1C_ACC/S2A_14TQL_20200921_0_L1C_ACC_QA.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QK/2020/8/9/LC08_L1TP_028032_20200809_20200917_02_T1_ACC/LC08_L1TP_028032_20200809_20200917_02_T1_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QK/2020/8/25/LC08_L1TP_028032_20200825_20200905_02_T1_ACC/LC08_L1TP_028032_20200825_20200905_02_T1_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/S2A_MSI_L1C_T14TQL_20201014_20201014_B02.img',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/S2A_MSI_L1C_T14TQL_20201014_20201014_B02.hdr',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/HLS.S10.T14TQL.2020288.T173227.v1.5.hdf',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/angle_output.hdf',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/S2A_14TQL_20201014_0_L1C_ACC_ac_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/S2A_14TQL_20201014_0_L1C_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-s2-acc/14/T/QL/2020/10/14/S2A_14TQL_20201014_0_L1C_ACC/S2A_14TQL_20201014_0_L1C_ACC_QA.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QM/2021/7/27/LC08_L1TP_028031_20210727_20210804_02_T1_ACC/LC08_L1TP_028031_20210727_20210804_02_T1_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QM/2021/8/28/LC08_L1TP_028031_20210828_20210901_02_T1_ACC/LC08_L1TP_028031_20210828_20210901_02_T1_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QK/2021/6/9/LC08_L1TP_028032_20210609_20210615_02_T1_ACC/LC08_L1TP_028032_20210609_20210615_02_T1_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QK/2021/7/27/LC08_L1TP_028032_20210727_20210804_02_T1_ACC/LC08_L1TP_028032_20210727_20210804_02_T1_ACC_cloud_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/14/T/QM/2021/4/22/LC08_L1TP_028031_20210422_20210430_02_T1_ACC/LC08_L1TP_028031_20210422_20210430_02_T1_ACC_cloud_mask.tif',
        >>> ]
        >>> for href in test_hrefs:
        ...     asset_name = None
        ...     asset_dict = {'href': href}
        ...     channels = _determine_s2_channels(asset_name, asset_dict)
        ...     print(f'channels={channels}')
    """
    asset_href = asset_dict['href']
    eo_band_names = [eob['name'] for eob in asset_dict.get('eo:bands', ())]
    # print(f'asset_href={asset_href}')
    # print(f'eo_band_names={eo_band_names}')

    if re.search(r'TCI\.(tiff?|jp2)$', asset_href, re.I):
        return SENSOR_CHANNEL_ALIAS['S2'].get('TCI', 'tci:3')
    elif re.search(r'PVI\.(tiff?|jp2)$', asset_href, re.I):
        return 'pvi:3'
    elif re.search(r'PVI\.(tiff?|jp2)$', asset_href, re.I):
        # PVI is preview image
        return 'pvi:3'
    elif re.search(r'AOT\.(tiff?|jp2)$', asset_href, re.I):
        # AOT is Aerosol Optical Thickness
        return 'aot'
    elif re.search(r'WVP\.(tiff?|jp2)$', asset_href, re.I):
        # https://sentinels.copernicus.eu/web/sentinel/user-guides/sentinel-2-msi/processing-levels/level-2
        # WV is Water Vapour
        return 'wvp'
    elif re.search(r'SCL\.(tiff?|jp2)$', asset_href, re.I):
        # https://sentinels.copernicus.eu/web/sentinel/user-guides/sentinel-2-msi/processing-levels/level-2
        # SCL Scene Classification map
        return 'scl'
    elif re.search(r'cloudmask\.(tiff?|jp2)$', asset_href, re.I):
        return 'cloudmask'
    elif re.search(r'SR_AEROSOL\.(tiff?|jp2)$', asset_href, re.I):
        return 'sr_aerosol_mask'
    elif len(eo_band_names) > 0:
        return '|'.join((SENSOR_CHANNEL_ALIAS['S2'].get(eobn, eobn)
                         for eobn in eo_band_names))
    elif m := re.search(r'(B\w{2})\.(tiff?|jp2)$', asset_href, re.I):  # NOQA
        return SENSOR_CHANNEL_ALIAS['S2'].get(m.group(1), m.group(1))
    else:
        href_path = ub.Path(asset_href)
        stem = href_path.stem
        known_suffixes = [
            'QA',
            'ac_mask',
            'cloud_mask',
        ]
        for suffix in known_suffixes:
            if stem.endswith('_' + suffix):
                return suffix
        return None


def _determine_l8_channels(asset_name, asset_dict):
    """
    Example:
        >>> from watch.cli.ta1_stac_to_kwcoco import *  # NOQA
        >>> from watch.cli.ta1_stac_to_kwcoco import _determine_l8_channels
        >>> test_hrefs = [
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/EG/2017/12/2/LC08_L1TP_114034_20171202_20200902_02_T1_ACC/LC08_L1TP_114034_20171202_20200902_02_T1_ACC_QA.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/EG/2017/12/2/LC08_L1TP_114034_20171202_20200902_02_T1_ACC/LC08_L1TP_114034_20171202_20200902_02_T1_ACC_TCI.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_ac_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_solar_zenith_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_solar_azimuth_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_view_zenith_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_view_azimuth_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_QA.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2017/9/20/LC08_L1TP_115034_20170920_20200903_02_T1_ACC/LC08_L1TP_115034_20170920_20200903_02_T1_ACC_TCI.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/EG/2017/9/13/LC08_L1TP_114034_20170913_20200903_02_T1_ACC/LC08_L1TP_114034_20170913_20200903_02_T1_ACC_ac_mask.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/EG/2017/9/13/LC08_L1TP_114034_20170913_20200903_02_T1_ACC/LC08_L1TP_114034_20170913_20200903_02_T1_ACC_solar_zenith_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/EG/2017/9/13/LC08_L1TP_114034_20170913_20200903_02_T1_ACC/LC08_L1TP_114034_20170913_20200903_02_T1_ACC_solar_azimuth_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/EG/2017/9/13/LC08_L1TP_114034_20170913_20200903_02_T1_ACC/LC08_L1TP_114034_20170913_20200903_02_T1_ACC_view_zenith_angle.tif',
        >>>     '/vsis3/smart-data-accenture/ta-1/ta1-ls-acc/52/S/DG/2020/6/8/LC08_L1TP_115034_20200608_20200824_02_T1_ACC/LC08_L1TP_115034_20200608_20200824_02_T1_ACC_cloud_mask.tif',
        >>> ]
        >>> for href in test_hrefs:
        ...     asset_name = None
        ...     asset_dict = {'href': href}
        ...     channels = _determine_l8_channels(asset_name, asset_dict)
        ...     print(f'channels={channels}')
    """
    asset_href = asset_dict['href']
    eo_band_names = []
    for eob in asset_dict.get('eo:bands', []):
        if isinstance(eob, dict):
            eo_band_names.append(eob['name'])
        elif isinstance(eob, str):
            eo_band_names.append(eob)
        else:
            raise TypeError(f'type(eob) = {type(eob)}')

    if len(eo_band_names) > 0:
        mapped_names = list(ub.unique([SENSOR_CHANNEL_ALIAS['L8'].get(eobn, eobn) for eobn in eo_band_names]))
        return '|'.join(mapped_names)
    elif re.search(r'cloudmask\.(tiff?|jp2)$', asset_href, re.I):
        return 'cloudmask'
    elif m := re.search(r'(QA_PIXEL|QA_RADSAT|QA_LINEAGE|SR_QA_AEROSOL)\.(tiff?|jp2)$',  # NOQA
                        asset_href, re.I):
        return m.group(1).lower()
    elif m := re.search(r'(B\w{1,2})\.(tiff?|jp2)$', asset_href, re.I):  # NOQA
        return SENSOR_CHANNEL_ALIAS['L8'].get(m.group(1), m.group(1))
    else:
        stem = ub.Path(asset_href).stem
        if stem.endswith('_TCI'):
            return 'tci:3'
        known_suffixes = [
            'QA',
            'ac_mask',
            'solar_zenith_angle',
            'view_zenith_angle',
            'solar_zenith_angle',
            'view_azimuth_angle',
            'view_zenith_angle',
            'solar_azimuth_angle',
            'solar_zenith_angle',
            'cloud_mask',
        ]
        for suffix in known_suffixes:
            if stem.endswith('_' + suffix):
                return suffix
        return None


def _determine_wv_channels(asset_name, asset_dict):
    from watch.utils import util_bands
    asset_href = asset_dict['href']

    eo_band_names = []
    for eob in asset_dict.get('eo:bands', []):
        if isinstance(eob, dict):
            eo_band_names.append(eob['name'])
        elif isinstance(eob, str):
            eo_band_names.append(eob)
        else:
            raise TypeError(f'type(eob) = {type(eob)}')

    if eo_band_names:
        channels = '|'.join(eo_band_names)
    else:
        from osgeo import gdal
        bands = gdal.Info(asset_href, format='json')['bands']

        # the channel names are the same for all WV, just the
        # center_wavelength is different so we can safely use this
        # info from WV2
        def _code(band_dicts):
            return '|'.join(b.get('common_name', b['name'])
                            for b in band_dicts)

        if len(bands) == 1:
            channels = _code(util_bands.WORLDVIEW2_PAN)
        elif len(bands) == 4:
            channels = _code(util_bands.WORLDVIEW2_MS4)
        elif len(bands) == 8:
            channels = _code(util_bands.WORLDVIEW2_MS8)
        else:
            raise Exception('unknown channel signature for WV')
    return channels


@profile
def make_coco_aux_from_stac_asset(asset_name,
                                  asset_dict,
                                  platform,
                                  name=None,
                                  force_affine=True,
                                  assume_relative=False,
                                  from_collated=False,
                                  populate_watch_fields=True,
                                  verbose=0):
    """
    Converts a single STAC asset into an "auxiliary" item / asset that will
    belong to a kwcoco image.
    """
    img = {}
    if name is not None:
        img['name'] = name

    asset_href = asset_dict['href']

    # Skip assets with metadata or thumbnail extensions
    if re.search(r'\.(txt|csv|json|xml)$', asset_href, re.I):
        if verbose:
            print(f'SKIP META asset: {asset_href}')
        return None

    if re.search(r'\.(vrt|jpe?g)$', asset_href, re.I):
        if verbose:
            print(f'SKIP THUMB asset: {asset_href}')
        return None

    if re.search(r'\.(img|hdr|hdf|imd)$', asset_href, re.I):
        if verbose:
            print(f'SKIP asset: {asset_href}')
        return None

    # HACK Skip common TCI (true color images) and PVI (preview images)
    # naming schemes
    if re.search(r'TCI\.jp2$', asset_href, re.I):
        if verbose:
            print(f'SKIP TCI asset: {asset_href}')
        return None
    if re.search(r'_PVI\.tif$', asset_href, re.I):
        if verbose:
            print(f'SKIP PVI asset: {asset_href}')
        return None

    if from_collated and platform in SUPPORTED_PLATFORMS:
        if verbose:
            print('Detected collated channels')
        channels = _determine_channels_collated(asset_name, asset_dict,
                                                platform)
    elif platform in SUPPORTED_COARSE_PLATFORMS['S2']:
        if verbose:
            print('Detected S2 channels')
        channels = _determine_s2_channels(asset_name, asset_dict)
    elif platform in SUPPORTED_COARSE_PLATFORMS['L8']:
        if verbose:
            print('Detected L8 channels')
        channels = _determine_l8_channels(asset_name, asset_dict)
    elif platform in SUPPORTED_COARSE_PLATFORMS['WV']:
        if verbose:
            print('Detected WV channels')
        channels = _determine_wv_channels(asset_name, asset_dict)
    elif platform in SUPPORTED_COARSE_PLATFORMS['WV1']:
        if verbose:
            print('Detected WV channels')
        channels = _determine_wv_channels(asset_name, asset_dict)
    else:
        raise NotImplementedError(
            "Unsupported platform '{}'".format(platform))

    # Hard-coded
    ignore_channels = [
        'solar_zenith_angle',
        'view_zenith_angle',
        'solar_zenith_angle',
        'view_azimuth_angle',
        'view_zenith_angle',
        'solar_azimuth_angle',
        'solar_zenith_angle',
        'tci:3',
    ]
    if channels is not None:
        if channels in ignore_channels:
            if verbose:
                print(f'SKIP ignored asset: {asset_href}')
            return None

    if channels is None:
        HACK_AWAY_SOME_WARNINGS = 1
        if HACK_AWAY_SOME_WARNINGS:
            # FIXME: parametarize or make robust
            IGNORE_SUFFIXES = (
                '_SAA.TIF', '_VZA.TIF', '_VAA.TIF', '_SZA.TIF'
            )
            if asset_href.endswith(IGNORE_SUFFIXES):
                return None
        # Collated output must always have eo:bands, so dont warn
        if not from_collated:
            print("* Warning * Couldn't determine channels for asset "
                  "at: '{}'. Asset will be ignored.".format(asset_href))
        elif verbose:
            print(f'SKIP Unknown asset: {asset_href}')
        return None

    if verbose:
        print(f'ADD asset: {asset_href}')

    if assume_relative:
        file_name = join(basename(dirname(asset_href)), basename(asset_href))
    else:
        file_name = asset_href

    img.update({
        'file_name': file_name,
        'channels': channels,
    })
    if populate_watch_fields:
        raise NotImplementedError('REMOVED: use coco_add_watch_feilds '
                                  'as a secondary step instead')
    return img


@profile
def _stac_item_to_kwcoco_image(stac_item,
                               assume_relative=False,
                               from_collated=False,
                               populate_watch_fields=False,
                               verbose=0):
    from watch.utils import util_time
    if populate_watch_fields:
        raise NotImplementedError('REMOVED: use coco_add_watch_feilds '
                                  'as a secondary step instead')
    stac_item_dict = stac_item.to_dict()

    platform = stac_item_dict['properties']['platform']
    if 'constellation' in stac_item_dict['properties']:
        if stac_item_dict['properties']['constellation'] == 'dove':
            platform = 'PD'

    if platform == 'DigitalGlobe' and stac_item_dict['properties']['mission'] == 'WV01':
        # Ensure we can differentiate between WV01 (which sometimes
        # has 'DigitalGlobe' as the 'platform') and other WV imagery
        platform = 'WV01'

    # Convet to standard case
    platform = PLATFORM_NORMALIZED_TO_PLATFORM_STANDARD_CASE.get(normalize_str(platform), platform)

    if platform not in SUPPORTED_PLATFORMS:
        print("* Warning * platform '{}' not supported, not adding to "
              "KWCOCO output!".format(platform))
        return None

    img = {
        'name': stac_item.id,
        'file_name': None,
    }
    auxiliary = []

    for asset_name, asset_dict in stac_item_dict.get('assets', {}).items():
        aux = make_coco_aux_from_stac_asset(
            asset_name,
            asset_dict,
            platform,
            force_affine=True,
            assume_relative=assume_relative,
            from_collated=from_collated,
            populate_watch_fields=populate_watch_fields,
            verbose=verbose,
        )
        if aux is not None:
            auxiliary.append(aux)

    if len(auxiliary) == 0:
        print("* Warning * Empty auxiliary assets for "
              "STAC Item '{}', skipping!".format(stac_item.id))
        return None

    if len(auxiliary) == 0:
        img['failed'] = stac_item

    img['auxiliary'] = auxiliary
    img['stac_properties'] = stac_item_dict['properties']
    date = stac_item_dict['properties']['datetime']
    date = util_time.coerce_datetime(date).isoformat()
    img['date_captured'] = date

    sensor_coarse = SENSOR_COARSE_MAPPING[platform]
    img['sensor_coarse'] = sensor_coarse

    return img


@profile
def ta1_stac_to_kwcoco(input_stac_catalog,
                       outpath,
                       assume_relative=False,
                       populate_watch_fields=False,
                       jobs=1,
                       from_collated=False,
                       ignore_duplicates=False,
                       verbose=1):

    if populate_watch_fields:
        raise NotImplementedError('REMOVED: use coco_add_watch_feilds '
                                  'as a secondary step instead')

    from watch.utils import util_parallel
    import pystac
    import json
    import kwcoco
    jobs = util_parallel.coerce_num_workers(jobs)

    if isinstance(input_stac_catalog, str):
        catalog = pystac.read_file(href=input_stac_catalog).full_copy()
    elif isinstance(input_stac_catalog, dict):
        catalog = pystac.Catalog.from_dict(input_stac_catalog).full_copy()
    else:
        catalog = input_stac_catalog.full_copy()

    outdir = dirname(outpath)
    os.makedirs(outdir, exist_ok=True)

    if jobs == 1:
        jobs = 0
    executor = ub.JobPool(mode='process', max_workers=jobs)

    all_items = [stac_item for stac_item in catalog.get_all_items()]

    if 1:
        if verbose > 5:
            # Printout one item per sensor
            sensor_to_one_item = {}
            for stac_item in all_items:
                # TODO: we can use this data to prepopulate the kwcoco file
                # so it takes far less time to field it.
                stac_dict = stac_item.to_dict()
                # stac_dict['geometry']
                sensor = stac_dict['properties'].get(
                    'constellation', stac_dict['properties'].get('platform', None))
                sensor_to_one_item[sensor] = stac_dict
            print('sensor_to_one_item = {}'.format(ub.urepr(sensor_to_one_item, nl=True)))

        # Sumamrize items before processing
        sensorchan_hist = ub.ddict(lambda: 0)
        sensorasset_hist = ub.ddict(lambda: 0)
        for stac_item in all_items:
            summary = summarize_stac_item(stac_item)
            sensor = summary['sensor']
            eo_bands = summary['eo_bands']
            asset_names = summary['asset_names']
            sensorchan = kwcoco.SensorChanSpec.coerce(f'{sensor}:' + '|'.join(eo_bands))
            sensorchan_hist[sensorchan.spec] += 1
            sensorasset = kwcoco.SensorChanSpec.coerce(f'{sensor}:' + '|'.join(sorted(asset_names)))
            sensorasset_hist[sensorasset.spec] += 1
            # TODO: stac_item['geometry'] - we can prepopulate geo information
            # stac_dict['properties']

        print('sensorchan_hist = {}'.format(ub.urepr(sensorchan_hist, nl=1)))
        print('sensorasset_hist = {}'.format(ub.urepr(sensorasset_hist, nl=1)))

    for stac_item in all_items:
        executor.submit(_stac_item_to_kwcoco_image, stac_item,
                        assume_relative=assume_relative,
                        from_collated=from_collated,
                        populate_watch_fields=populate_watch_fields,
                        verbose=verbose > 1)

    output_dset = kwcoco.CocoDataset()
    output_dset.fpath = outpath

    # TODO: Should make this name the MGRS tile
    for job in executor.as_completed(desc='collect jobs', progkw={'verbose': verbose}):
        kwcoco_img = job.result()
        if kwcoco_img is not None:
            # Ignore iamges with 0 auxiliary items
            if len(kwcoco_img.get('auxiliary', [])) == 0:
                print('Failed kwcoco_img = {}'.format(ub.urepr(kwcoco_img, nl=1)))
                continue
            try:
                output_dset.add_image(**kwcoco_img)
            except kwcoco.exceptions.DuplicateAddError:
                if not ignore_duplicates:
                    print(ub.paragraph(
                        '''
                        Error encountered duplicate item. Debugging duplicates.
                        Did you append to the same input list multiple times?
                        '''))
                    raise

    with open(outpath, 'w') as f:
        json.dump(output_dset.dataset, f, indent=2)

    print('Wrote: {}'.format(outpath))

    return output_dset


def summarize_stac_item(stac_item):
    # TODO: we can use this data to prepopulate the kwcoco file
    # so it takes far less time to field it.
    stac_dict = stac_item.to_dict()
    # stac_dict['geometry']
    sensor = stac_dict['properties'].get(
        'constellation', stac_dict['properties'].get('platform', None))
    # proc_level = stac_dict['landsat:correction']
    asset_names = stac_dict['assets'].keys()

    eo_bands = []
    for asset_name, asset_item in stac_dict['assets'].items():
        if 'roles' in asset_item and 'data' in asset_item['roles']:
            if 'eo:bands' in asset_item:
                for eo_band in asset_item['eo:bands']:
                    if isinstance(eo_band, dict):
                        if 'common_name' in eo_band:
                            eo_bands.append(eo_band['common_name'])
                        elif 'name' in eo_band:
                            eo_bands.append(eo_band['name'])
                        else:
                            raise AssertionError
                    elif isinstance(eo_band, str):
                        eo_bands.append(eo_band)
    eo_bands = list(ub.unique(eo_bands))
    eo_cloud_cover = stac_dict['properties'].get('eo:cloud_cover', None)

    summary = {
        'sensor': sensor,
        'asset_names': asset_names,
        'eo_bands': eo_bands,
        'eo_cloud_cover': eo_cloud_cover,
        'datetime': stac_item.get_datetime(),
    }
    return summary


if __name__ == "__main__":
    sys.exit(main())

"""
Test:
    import pandas as pd

    freq = ub.ddict(list)
    for img in dset.dataset['images']:
        sensor = img['sensor_coarse']
        freq[sensor].append(len(img['auxiliary']))

    for sensor, auxfreq in freq.items():
        print('sensor = {!r}'.format(sensor))
        print(ub.dict_hist(auxfreq))
        pass


"""



def _dev_1darray_sample():
    """
    What are other options to sample roughly uniformly from clumpy data?

    References:
        https://stats.stackexchange.com/questions/122668/is-there-a-measure-of-evenness-of-spread
    """
    import numpy as np
    import ubelt as ub
    # The idea is that we are given cluttered datetimes
    from watch.utils.util_time import coerce_timedelta, coerce_datetime
    from watch.tasks.fusion.datamodules.temporal_sampling.plots import plot_temporal_sample_indices
    # Generate a "clumpy" sample

    def demo_clumpy_data(N, rng):
        uniform = np.linspace(0, 1, N)
        noise = rng.randn(N) / N
        initial = kwarray.normalize(uniform + noise)
        num_clumps = int(N // 10)
        # ranks = initial.argsort()
        ranks = rng.rand(N).argsort()
        clump_size = 7
        clump_indexes = ranks[0:num_clumps]
        # eaten_indexes = ranks[num_clumps:num_clumps + num_clumps * clump_size]
        remain_indexes = ranks[num_clumps + num_clumps * clump_size:]
        clump_seeds = uniform[clump_indexes]
        clump_members = (clump_seeds[:, None] + rng.randn(num_clumps, clump_size) / N).ravel()
        clumpy = kwarray.normalize(np.r_[initial[clump_indexes], initial[remain_indexes], clump_members])
        clumpy.sort()
        return clumpy
    import kwarray
    N = 100
    rng = kwarray.ensure_rng()
    clumpy = demo_clumpy_data(N, rng)
    start_time = coerce_datetime('now').timestamp()
    obs_time = coerce_timedelta('10years').total_seconds()
    # time_span = N * sample_delta
    # delta = time_span.total_seconds()
    unixtimes = (clumpy * obs_time) + start_time
    # time_span
    import kwplot
    plt = kwplot.autoplt()
    plt.plot(unixtimes, 'o')

    k_sizes = [3, 5, 10, 20]
    all_idxs = np.arange(len(unixtimes))
    with ub.Timer('kmeans'):
        sample_idxs = [all_idxs]
        for k in k_sizes:
            from sklearn.cluster import KMeans
            km = KMeans(n_clusters=k)
            km = km.fit(unixtimes[:, None])
            labels = km.predict(unixtimes[:, None])
            # idx_to_cluster = kwarray.group_items(unixtimes, labels)
            cx_to_sxs = ub.dzip(*kwarray.group_indices(labels))
            km_sample_idxs = []
            for cx, sxs in cx_to_sxs.items():
                cluster = unixtimes[sxs]
                midx = sxs[np.abs(cluster - km.cluster_centers_[cx]).argmin()]
                km_sample_idxs.append(midx)
            km_sample_idxs = np.array(km_sample_idxs)
            sample_idxs.append(km_sample_idxs)
        # plot_temporal_sample_indices(sample_idxs, unixtimes)

    # from sklearn.cluster import MeanShift
    # ms = MeanShift()
    # ms.fit(unixtimes[:, None])
    # Uniform bins method
    with ub.Timer('1step'):
        norm_data =  kwarray.normalize(unixtimes)
        for k in k_sizes:
            centers = np.linspace(0, 1, k, endpoint=0) + (1 / (k * 2))
            labels = np.abs(norm_data[None, :] - centers[:, None]).argmin(axis=0)
            cx_to_sxs = ub.dzip(*kwarray.group_indices(labels))
            km_sample_idxs = []
            for cx, sxs in cx_to_sxs.items():
                cluster = unixtimes[sxs]
                midx = sxs[np.abs(cluster - km.cluster_centers_[cx]).argmin()]
                km_sample_idxs.append(midx)
            km_sample_idxs = np.array(km_sample_idxs)
            sample_idxs.append(km_sample_idxs)

    plot_temporal_sample_indices(sample_idxs, unixtimes)


def test_time_strategy():
    # Basic overview demo of the algorithm
    from watch.tasks.fusion.datamodules.temporal_sampling import TimeWindowSampler
    import watch
    dset = watch.coerce_kwcoco('watch-msi', geodata=True, dates=True, num_frames=128, image_size=(32, 32))
    vidid = dset.dataset['videos'][0]['id']
    self = TimeWindowSampler.from_coco_video(
        dset, vidid,
        time_window=11,
        affinity_type='soft2', time_span='8m', update_rule='distribute',
    )
    # xdoctest: +REQUIRES(--show)
    import kwplot
    kwplot.autosns()

    # xdoctest: +REQUIRES(env:SMART_DATA_DVC_DPATH)
    # from watch.tasks.fusion.datamodules.temporal_sampling import *  # NOQA
    import watch
    data_dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    coco_fpath = data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.json'
    dset = watch.coerce_kwcoco(coco_fpath)
    vidid = dset.dataset['videos'][0]['id']
    self = TimeWindowSampler.from_coco_video(
        dset, vidid,
        time_window=11,
        affinity_type='hardish3', time_span='3m', update_rule='pairwise+distribute', determenistic=True
    )

    # from scipy.special import expit
    from watch.tasks.fusion.datamodules.temporal_sampling.utils import coerce_time_kernel
    from watch.utils.util_time import coerce_timedelta
    # import kwarray
    import numpy as np
    pattern = '-1y,-60d,-30d,-1d,0,1d,30d,60d,1y'
    kernel = coerce_time_kernel(pattern)

    delta_diff = (self.unixtimes[:, None] - self.unixtimes[None, :])
    kwplot.autoplt().imshow(delta_diff, cmap='coolwarm')

    [coerce_timedelta(d) for d in delta_diff[0]][0:10]

    diff = np.abs((delta_diff - kernel[:, None, None]))
    sdiff = diff - diff.min(axis=0)[None, :, :]
    energy = sdiff.mean()
    kwplot.autoplt().imshow(sdiff.mean(axis=0), cmap='magma')

    s = 1 / sdiff.mean(axis=0)
    flags = np.isinf(s)
    s[flags] = 0
    s = (s / s.max())
    s[flags] = 1
    kwplot.autoplt().imshow(s, cmap='magma')
    distance_weight = s

    # delta_diff01 = kwarray.normalize(delta_diff)
    # kernel01 = kwarray.normalize(kernel, min_val=delta_diff.min(), max_val=delta_diff.max())

    sensor_value = {
        'WV': 10,
        'WV1': 9,
        'S2': 1,
        'PD': 7,
        'L8': 0.3,
        'sensor1': 11,
        'sensor2': 7,
        'sensor3': 5,
        'sensor4': 3,
    }
    import ubelt as ub
    values = np.array(list(ub.take(sensor_value, self.sensors, default=1))).astype(float)
    values /= values.max()
    sensor_weight = np.sqrt(values[:, None] * values[None, :])
    # distance = (1 - np.abs((delta_diff01 - kernel01[:, None, None])).min(axis=0))

    # kwplot.autoplt().imshow(distance, cmap='plasma')
    # kwplot.autoplt().imshow(distance ** 4, cmap='plasma')

    energy = distance_weight * sensor_weight
    # energy = distance_weight
    energy = (energy / np.diag(energy)[:, None]).clip(None, 1)
    energy.max(axis=0)

    self.affinity = self.affinity * energy

    kwplot.autoplt().imshow(energy, cmap='plasma')

    import kwplot
    plt = kwplot.autoplt()
    self.show_summary(samples_per_frame=11, fnum=1)
    self.show_procedure(fnum=4)
    plt.subplots_adjust(top=0.9)


def _format_xaxis_as_timedelta(ax):
    """
    Add to util_kwplot
    """
    import datetime as datetime_mod

    def timeTicks(x, pos):
        d = datetime_mod.timedelta(seconds=x)
        return str(d.days)  # + ' days'
    import matplotlib as mpl
    formatter = mpl.ticker.FuncFormatter(timeTicks)
    ax.xaxis.set_major_formatter(formatter)


def _visualize_temporal_sampling():
    from watch.utils import util_time
    import numpy as np
    from watch.tasks.fusion.datamodules.temporal_sampling.utils import coerce_time_kernel
    import kwarray

    # time_kernel = coerce_time_kernel('-1y,-6m,-3m,0,3m,6m,1y')
    time_kernel = coerce_time_kernel('-1y,-3m,0,3m,1y')
    time_range = util_time.coerce_timedelta('4y')

    rng = kwarray.ensure_rng()
    relative_unixtimes = rng.rand(10) * time_range.total_seconds()
    idx = ((relative_unixtimes - (time_range.total_seconds() / 2)) ** 2).argmin()
    mid = relative_unixtimes[idx]
    relative_unixtimes = relative_unixtimes - mid

    # time_kernel = coerce_time_kernel('-1H,-5M,0,5M,1H')
    # relative_unixtimes = coerce_time_kernel('-90M,-70M,-50M,0,1sec,10S,30M')
    # relative_unixtimes = coerce_time_kernel('-90M,-70M,-50M,-20M,-10M,0,1sec,10S,30M,57M,87M')

    from watch.tasks.fusion.datamodules.temporal_sampling.affinity import make_soft_mask
    kernel_masks, kernel_attrs = make_soft_mask(time_kernel, relative_unixtimes)

    min_t = min(kattr['left'] for kattr in kernel_attrs)
    max_t = max(kattr['right'] for kattr in kernel_attrs)

    min_t = min(min_t, relative_unixtimes[0])
    max_t = max(max_t, relative_unixtimes[-1])

    import kwplot
    from watch.utils import util_kwplot
    plt = kwplot.autoplt()
    import kwimage
    kwplot.close_figures()
    kwplot.figure(fnum=1, doclf=1)
    kernel_color = kwimage.Color.coerce('kitware_green').as01()
    obs_color = kwimage.Color.coerce('kitware_blue').as01()

    kwplot.figure(fnum=1, pnum=(1, 1, 1))

    kwplot.phantom_legend({
        'ideal sample': kernel_color,
        'available observation': obs_color,
    })

    for kattr in kernel_attrs:
        rv = kattr['rv']
        xs = np.linspace(min_t, max_t, 1000)
        ys = rv.pdf(xs)
        kattr['_our_norm'] = ys.sum()
        ys_norm = ys / ys.sum()
        plt.plot(xs, ys_norm)

    ax = plt.gca()
    # ax.set_ylim(0, 1)
    ax.set_xlabel('relative time (days)')
    ax.set_ylabel('sample probability')
    # ax.set_title('ideal sample location')
    ax.set_yticks([])

    obs_line_segments = []
    for x in relative_unixtimes:
        y = 0
        for kattr in kernel_attrs:
            rv = kattr['rv']
            y = max(y, rv.pdf(x) / kattr['_our_norm'])
        obs_line_segments.append([x, y])
    for x, y in obs_line_segments:
        plt.plot([x, x], [0, y], '-', color=obs_color)

    kern_line_segments = []
    for x in time_kernel:
        y = 0
        for kattr in kernel_attrs:
            rv = kattr['rv']
            y = max(y, rv.pdf(x) / kattr['_our_norm'])
        kern_line_segments.append([x, y])
    for x, y in kern_line_segments:
        plt.plot([x, x], [0, y], '--', color=kernel_color)

    # plt.plot(time_kernel, [0] * len(time_kernel), '-o', color=kernel_color, label='ideal frame location')

    kwplot.phantom_legend(label_to_attrs={
        'ideal sample': {'color': kernel_color, 'linestyle': '--'},
        'available observation': {'color': obs_color},
    })

    _format_xaxis_as_timedelta(ax)
    plt.subplots_adjust(top=0.99, bottom=0.1, hspace=.3, left=0.1)
    fig = plt.gcf()
    fig.set_size_inches(np.array([4, 3]) * 1.5)
    fig.tight_layout()
    finalizer = util_kwplot.FigureFinalizer()
    finalizer(fig, 'time_sampling_example.png')



import torch
import torch.nn as nn


class MatED(nn.Module):
    def __init__(self, task_mode, encoder, discretizer, decoder, attention_block=None, encoder_out_layer="layer1"):
        super(MatED, self).__init__()

        self.task_mode = task_mode
        self.encoder = encoder
        self.decoder = decoder
        self.discretizer = discretizer
        self.attention_block = attention_block

    def encode_mat_feats(self, frame):
        # frame: [batch_size, n_channels, height, width]

        # Get continuous features
        if self.attention_block:
            feats = self.encoder(frame)  # [batch_size, n_channels, height, width]
            feats = self.attention_block(feats)["layer1"]
        else:
            feats = self.encoder(frame)["layer1"]  # [batch_size, n_channels, height, width]

        # Use learned dictionary to get discrete features.
        d_feats, codewords, kld_loss, mat_confs = self.discretizer(feats)

        return d_feats, mat_confs

    def mat_id_to_image(self, mat_id):
        # mat_id: [height, width]

        # Get material features from mat IDs.
        mat_feat = self.discretizer.mat_id_to_mat_feat(mat_id)

        # Resize material features.
        # from: [height, width, feat_dim]
        # to: [1, feat_dim, height, width]
        mat_feat = mat_feat.permute(2, 0, 1).unsqueeze(0)

        # Generate image from material features.
        image = self.decoder(mat_feat)

        return image

    def forward(self, data):
        output = {}

        # Get image.
        video = data["video"]  # [batch_size, 1, n_channels, height, width]
        ## Get only frame from video.
        frame = video[:, 0]

        # Get continuous features
        if self.attention_block:
            feats = self.encoder(frame)  # [batch_size, n_channels, height, width]
            feats = self.attention_block(feats)["layer1"]
        else:
            feats = self.encoder(frame)["layer1"]  # [batch_size, n_channels, height, width]

        if self.discretizer:
            # Use learned dictionary to get discrete features.
            d_feats, codewords, kld_loss, mat_confs = self.discretizer(feats)
            output["kld_loss"] = kld_loss
            output["mat_ids"] = codewords
            output["mat_conf"] = mat_confs
        else:
            d_feats = feats

        # Generate prediction with decoder.
        pred = self.decoder(d_feats)

        output[self.task_mode] = pred.unsqueeze(1)

        return output


class MTMatED(nn.Module):
    def __init__(self, task_mode, encoder, discretizer, decoder, attention_block=None, encoder_out_layer="layer1"):
        super(MTMatED, self).__init__()

        self.task_mode = task_mode
        self.encoder = encoder
        self.decoder = decoder
        self.discretizer = discretizer
        self.attention_block = attention_block

    def encode_mat_feats(self, frame):
        # frame: [batch_size, n_channels, height, width]

        # Get continuous features
        if self.attention_block:
            feats = self.encoder(frame)  # [batch_size, n_channels, height, width]
            feats = self.attention_block(feats)["layer1"]
        else:
            feats = self.encoder(frame)["layer1"]  # [batch_size, n_channels, height, width]

        # Use learned dictionary to get discrete features.
        d_feats, codewords, kld_loss, mat_confs = self.discretizer(feats)

        return d_feats, mat_confs

    def encode(self, frame):
        # Get continuous features
        if self.attention_block:
            feats = self.encoder(frame)  # [batch_size, n_channels, height, width]
            feats = self.attention_block(feats)["layer1"]
        else:
            feats = self.encoder(frame)["layer1"]  # [batch_size, n_channels, height, width]
        return feats

    def mat_id_to_image(self, mat_id):
        # mat_id: [height, width]

        # Get material features from mat IDs.
        mat_feat = self.discretizer.mat_id_to_mat_feat(mat_id)

        # Resize material features.
        # from: [height, width, feat_dim]
        # to: [1, feat_dim, height, width]
        mat_feat = mat_feat.permute(2, 0, 1).unsqueeze(0)

        # Generate image from material features.
        image = self.decoder(mat_feat)

        return image

    def forward(self, data):
        output = {}

        # Get image.
        video = data["video"]  # [batch_size, n_frames, n_channels, height, width]
        n_frames = video.shape[1]

        preds, mat_masks, mat_confs = [], [], []
        kld_losses = 0
        for i in range(n_frames):
            frame = video[:, i]
            feats = self.encode(frame)

            if self.discretizer:
                # Use learned dictionary to get discrete features.
                d_feats, codewords, kld_loss, mat_conf = self.discretizer(feats)
                kld_losses += kld_loss
                mat_masks.append(codewords)
                mat_confs.append(mat_conf)
            else:
                d_feats = feats

            # Generate prediction with decoder.
            pred = self.decoder(d_feats)
            preds.append(pred)

        output = {}
        if self.discretizer:
            output["kld_loss"] = kld_losses / n_frames
            output["mat_ids"] = torch.stack(mat_masks, dim=1)
            output["mat_conf"] = torch.stack(mat_confs, dim=1)

            # Compute material consistency loss.
            output["mat_consistency_loss"] = (
                (output["mat_conf"].mean(dim=2, keepdim=True) - output["mat_conf"]).std(dim=1).sum(1).mean()
            )

        output[self.task_mode] = torch.stack(preds, dim=1)
        return output



# flake8: noqa

import comet_ml
import time
from watch.tasks.rutgers_material_seg.models.canny_edge import CannyFilter
from skimage.filters import threshold_otsu as otsu
from fast_pytorch_kmeans import KMeans
from watch.tasks.rutgers_material_seg.models.losses import SupConLoss, simCLR_loss, QuadrupletLoss, SoftCE
from watch.tasks.rutgers_material_seg.models.supcon import SupConResNet
from watch.tasks.rutgers_material_seg.datasets import build_dataset
from watch.tasks.rutgers_material_seg.datasets.iarpa_contrastive_dataset import SequenceDataset
from watch.tasks.rutgers_material_seg.models import build_model
import watch.tasks.rutgers_material_seg.utils.visualization as visualization
import watch.tasks.rutgers_material_seg.utils.eval_utils as eval_utils
import watch.tasks.rutgers_material_seg.utils.utils as utils
from pytorch_metric_learning.distances import SNRDistance
from pytorch_metric_learning import losses
import torchvision.transforms.functional as FT
from torchvision import transforms
from tqdm import tqdm
from torch import nn
from scipy import ndimage
import torch.nn.functional as F
import torch.optim as optim
import ubelt as ub
import numpy as np
import matplotlib.pyplot as plt
import ndsampler
import kwimage
import kwcoco
import random
import math
import yaml
import warnings
import datetime
import torch
import cv2
import gc
import matplotlib
matplotlib.use('Agg')

import sys
import os
debug_mode = True
if debug_mode:
    from pympler import muppy, summary
current_path = os.getcwd().split("/")

torch.backends.cudnn.enabled = False
torch.backends.cudnn.deterministic = True
torch.set_printoptions(precision=3, sci_mode=False)
np.set_printoptions(precision=3, suppress=True)

mean_std = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])


class Trainer(object):
    def __init__(self, model: object, train_loader: torch.utils.data.DataLoader,
                 val_loader: torch.utils.data.DataLoader, epochs: int,
                 optimizer: object, scheduler: object,
                 test_loader: torch.utils.data.DataLoader = None,
                 test_with_full_supervision: int = 0) -> None:
        """trainer class

        Args:
            model (object): trained or untrained model
            train_loader (torch.utils.data.DataLader): loader with training data
            val_loader (torch.utils.data.DataLader): loader with validation data
            epochs (int): number of epochs
            optimizer (object): optimizer to train with
            scheduler (object): scheduler to train with
            test_loader (torch.utils.data.DataLader, optional): loader with testing data. Defaults to None.
            test_with_full_supervision (int, optional): should full supervision be used. Defaults to 0.
        """

        self.model = model
        self.use_crf = config['evaluation']['use_crf']
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.epochs = epochs
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.class_weights = torch.Tensor(config['data']['weights']).float().to(device)
        self.contrastive_loss = SupConLoss()
        self.quadrupletloss = QuadrupletLoss()
        self.triplet_margit_loss_snr = losses.TripletMarginLoss(margin=0.05, swap=False, smooth_loss=False, triplets_per_anchor="all", distance=SNRDistance())
        # self.k = config['training']['out_features_dim']
        self.k = config['data']['num_classes']
        self.kmeans = KMeans(n_clusters=self.k, mode='euclidean', verbose=0, minibatch=None)
        self.max_label = config['data']['num_classes']
        self.all_crops_params = [tuple([i, j, config['data']['window_size'], config['data']['window_size']]) for i in range(config['data']['window_size'], config['data']
                                                                                                                            ['image_size'] - config['data']['window_size']) for j in range(config['data']['window_size'], config['data']['image_size'] - config['data']['window_size'])]
        self.inference_all_crops_params = [tuple([i, j, config['evaluation']['inference_window'], config['evaluation']['inference_window']]) for i in range(0, config['data']['image_size']) for j in range(0, config['data']['image_size'])]
        self.all_crops_params_np = np.array(self.all_crops_params)
        self.change_threshold = 0.2
        # print(self.all_crops_params_np)
        if test_loader is not None:
            self.test_loader = test_loader
            self.test_with_full_supervision = test_with_full_supervision

        self.train_second_transform = transforms.Compose([
            transforms.RandomHorizontalFlip(),
            transforms.ColorJitter(0.4, 0.4, 0.4, 0.1),
            transforms.RandomGrayscale(p=0.2),
        ])
        self.crop_size = (config['data']['window_size'],
                          config['data']['window_size'])

        self.cmap = visualization.n_distinguishable_colors(nlabels=self.max_label,
                                                           first_color_black=True, last_color_black=True,
                                                           bg_alpha=config['visualization']['bg_alpha'],
                                                           fg_alpha=config['visualization']['fg_alpha'])

    def train(self, epoch: int, cometml_experiemnt: object) -> float:
        """training single epoch

        Args:
            epoch (int): number of epoch
            cometml_experiemnt (object): comet ml experiment to log the epoch

        Returns:
            float: training loss of that epoch
        """
        total_loss = 0
        total_loss_seg = 0
        preds, targets = [], []

        max_run_network_time = 0
        max_backprop_time = 0
        max_logging_time = 0

        self.model.train()
        print(f"starting epoch {epoch}")
        loader_size = len(self.train_loader)
        if config['visualization']['train_visualization_divisor'] >= loader_size:
            config['visualization']['train_visualization_divisor'] = loader_size
        iter_visualization = loader_size // config['visualization']['train_visualization_divisor']
        pbar = tqdm(enumerate(self.train_loader), total=len(self.train_loader))
        batch_index_to_show = config['visualization']['batch_index_to_show']
        for batch_index, batch in pbar:
            random_crop = transforms.RandomCrop(self.crop_size)
            outputs = batch
            image, mask = outputs['inputs']['im'].data[0], batch['label']['class_masks'].data[0]
            # print(f"image min:{image.min()}, max:{image.max()}")
            image_name = f"{str(outputs['tr'].data[0][batch_index_to_show]['gids'])}_{str(outputs['tr'].data[0][batch_index_to_show]['slices'])}"
            # original_width, original_height = outputs['tr'].data[0][batch_index_to_show]['space_dims']
            mask = torch.stack(mask)
            mask = mask.long().squeeze(1)
            start = time.time()
            mask[mask == -1] = 0
            mask[mask >= 4] = 0
            # print(torch.unique(mask))

            bs, c, t, h, w = image.shape

            class_to_show = max(0, torch.unique(mask)[-1] - 1)
            image = image.to(device).squeeze(2)
            mask = mask.to(device)

            # image = utils.stad_image(image)
            # image2 = utils.stad_image(image2)
            image = F.normalize(image, dim=1, p=1)

            output = self.model(image)  # [B,22,150,150]
            # print(output.shape)
            # exit()
            loss = 50 * F.cross_entropy(output,
                                    mask,
                                    weight=self.class_weights,
                                    # ignore_index=0,
                                    reduction="mean")

            run_network_time = time.time() - start

            start = time.time()
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            total_loss_seg += loss.item()

            backprop_time = time.time() - start

            masks = F.softmax(output, dim=1)
            pred1 = masks.max(1)[1].cpu().detach()  # .numpy()

            total_loss += loss.item()

            start = time.time()

            preds.append(pred1)
            targets.append(mask.cpu())  # .numpy())

            if config['visualization']['train_visualizer']:
                if (epoch) % config['visualization']['visualize_training_every'] == 0:
                    if (batch_index % iter_visualization) == 0:
                        figure = plt.figure(figsize=(config['visualization']['fig_size'], config['visualization']['fig_size']),
                                            dpi=config['visualization']['dpi'])
                        ax1 = figure.add_subplot(1, 4, 1)
                        ax2 = figure.add_subplot(1, 4, 2)
                        ax3 = figure.add_subplot(1, 4, 3)
                        ax4 = figure.add_subplot(1, 4, 4)
                        # ax5 = figure.add_subplot(3, 4, 5)
                        # ax6 = figure.add_subplot(3, 4, 6)
                        # ax7 = figure.add_subplot(3, 4, 7)
                        # ax8 = figure.add_subplot(3, 4, 8)
                        # ax9 = figure.add_subplot(3, 4, 9)
                        # ax10 = figure.add_subplot(3, 4, 10)
                        # ax11 = figure.add_subplot(3, 4, 11)
                        # ax12 = figure.add_subplot(3, 4, 12)

                        cmap_gradients = plt.cm.get_cmap('jet')
                        # image_show = np.transpose(image1.cpu().detach().numpy()[batch_index_to_show,:,:,:],(1,2,0))[:,:1:4,:3]
                        image_show1 = np.transpose(image.cpu().detach().numpy()[batch_index_to_show, :, :, :], (1, 2, 0))[:, :, :3]
                        image_show1 = np.flip(image_show1, axis=2)

                        image_show1 = (image_show1 - image_show1.min()) / (image_show1.max() - image_show1.min())
                        gt_mask_show1 = mask.cpu().detach()[batch_index_to_show, :, :].numpy().squeeze()

                        # histograms_intersection_show = (histograms_intersection_show - histograms_intersection_show.min())/(histograms_intersection_show.max() - histograms_intersection_show.min())

                        pred1_show = masks.max(1)[1].cpu().detach().numpy()[batch_index_to_show, :, :]
                        # pred2_show = masks2.max(1)[1].cpu().detach().numpy()[batch_index_to_show, :, :]

                        # histc_fp_tp_fn_prediction_mask = gt_mask_show1 + (2*histc_int_change_feats_pred_show)
                        pred1_fp_tp_fn_prediction_mask = gt_mask_show1 + (2 * pred1_show)

                        uniques_str = f"unique in preds: {np.unique(pred1_show)}, masks: {np.unique(gt_mask_show1)}"

                        classes_in_gt = np.unique(gt_mask_show1)
                        ax1.imshow(image_show1)

                        ax2.imshow(image_show1)
                        ax2.imshow(gt_mask_show1, cmap=self.cmap, vmin=0, vmax=self.max_label)

                        masks_show = masks.cpu().detach().numpy()[batch_index_to_show, 2, :, :]
                        ax3.imshow(masks_show)

                        ax4.imshow(pred1_show, cmap=self.cmap, vmin=0, vmax=self.max_label)
                        # ax11.imshow(histc_fp_tp_fn_prediction_mask, cmap=self.cmap, vmin=0, vmax=self.max_label)

                        # ax12.imshow(vw_dis_fp_tp_fn_prediction_mask, cmap=self.cmap, vmin=0, vmax=self.max_label)

                        ax1.axis('off')
                        ax2.axis('off')
                        ax3.axis('off')
                        ax4.axis('off')
                        # ax9.axis('off')

                        if config['visualization']['titles']:
                            ax1.set_title(f"Input Image 1", fontsize=config['visualization']['font_size'])
                            ax2.set_title(f"Change GT Mask overlaid", fontsize=config['visualization']['font_size'])
                            ax3.set_title(f"Features", fontsize=config['visualization']['font_size'])
                            ax4.set_title(f"Prediction {uniques_str}", fontsize=config['visualization']['font_size'])
                            # ax5.set_title(f"l1_patched_diff_change_features_show", fontsize=config['visualization']['font_size'])
                            # ax6.set_title(f"l2_patched_diff_change_features_show", fontsize=config['visualization']['font_size'])
                            # ax7.set_title(f"histograms_intersection_show", fontsize=config['visualization']['font_size'])
                            # ax8.set_title(f"dictionary2_show", fontsize=config['visualization']['font_size'])
                            # ax9.set_title(f"l1_fp_tp_fn_prediction_mask", fontsize=config['visualization']['font_size'])
                            # ax10.set_title(f"l2_fp_tp_fn_prediction_mask", fontsize=config['visualization']['font_size'])
                            # ax11.set_title(f"histc_fp_tp_fn_prediction_mask", fontsize=config['visualization']['font_size'])
                            # figure.suptitle(
                            #     f"Epoch: {epoch+1}\nGT labels for classification: {classes_in_gt}, \nunique in change predictions: {np.unique(change_detection_show)}\nunique in predictions1: {np.unique(logits_show1)}", fontsize=config['visualization']['font_size'])

                        # cometml_experiemnt.log_figure(figure_name=f"Training, image name: {image_name}, epoch: {epoch}, classes in gt: {classes_in_gt}, classifier predictions: {labels_predicted_indices}",figure=figure)
                        cometml_experiemnt.log_figure(figure_name=f"Training, image name: {image_name}", figure=figure)
                        # figure.tight_layout()

                        if config['visualization']['train_imshow']:
                            plt.show()

                        figure.clear()
                        figure.clf()
                        plt.cla()
                        plt.clf()
                        plt.close('all')
                        plt.close(figure)
                        gc.collect()

            logging_time = time.time() - start

            if run_network_time > max_run_network_time:
                max_run_network_time = run_network_time
            if backprop_time > max_backprop_time:
                max_backprop_time = backprop_time
            if logging_time > max_logging_time:
                max_logging_time = logging_time

            pbar.set_description(
                f"(timing, secs) run_network: {run_network_time:0.3f}({max_run_network_time:0.3f}), backprob: {backprop_time:0.3f}({max_backprop_time:0.3f}), log: {logging_time:0.3f}({max_logging_time:0.3f})")

        mean_iou, precision, recall = eval_utils.compute_jaccard(preds, targets, num_classes=config['data']['num_classes'])

        # hist_mean_iou, hist_precision, hist_recall = eval_utils.compute_jaccard(histogram_distance, targets, num_classes=2)
        # l1_mean_iou, l1_precision, l1_recall = eval_utils.compute_jaccard(l1_dist, targets, num_classes=2)
        # l2_mean_iou, l2_precision, l2_recall = eval_utils.compute_jaccard(l2_dist, targets, num_classes=2)

        # l1_precision = np.array(l1_precision)
        # l1_recall = np.array(l1_recall)
        # l1_f1 = 2 * (l1_precision * l1_recall) / (l1_precision + l1_recall)

        # l2_precision = np.array(l2_precision)
        # l2_recall = np.array(l2_recall)
        # l2_f1 = 2 * (l2_precision * l2_recall) / (l2_precision + l2_recall)

        # hist_precision = np.array(hist_precision)
        # hist_recall = np.array(hist_recall)
        # hist_f1 = 2 * (hist_precision * hist_recall) / (hist_precision + hist_recall)

        mean_iou = np.array(mean_iou)
        precision = np.array(precision)
        recall = np.array(recall)
        classwise_f1_score = 2 * (precision * recall) / (precision + recall + 0.00000000001)

        mean_precision = precision.mean()
        mean_recall = recall.mean()
        overall_miou = mean_iou.mean()
        mean_f1_score = classwise_f1_score.mean()
        mean_f1_score_from_r_p = 2 * (mean_precision * mean_recall) / (mean_precision + mean_recall + 0.00000000001)

        # overall_miou = sum(mean_iou)/len(mean_iou)
        # print(f"Training class-wise mIoU value: \n{mean_iou} \noverall mIoU: {overall_miou}")
        # print(f"Training class-wise Precision value: \n{precision} \noverall Precision: {mean_precision}")
        # print(f"Training class-wise Recall value: \n{recall} \noverall Recall: {mean_recall}")
        # print(f"Training overall F1 Score: {mean_f1_score}")

        # cometml_experiemnt.log_metric("Training Loss", total_loss, epoch=epoch+1)
        # cometml_experiemnt.log_metric("Segmentation Loss", total_loss_seg, epoch=epoch+1)
        # cometml_experiemnt.log_metric("Training mIoU", overall_miou, epoch=epoch+1)
        cometml_experiemnt.log_metric("Training mean f1_score_rp", mean_f1_score_from_r_p, epoch=epoch + 1)
        cometml_experiemnt.log_metric("Training mean_f1_score", mean_f1_score, epoch=epoch + 1)
        cometml_experiemnt.log_metric("Training mean_precision", mean_precision, epoch=epoch + 1)
        cometml_experiemnt.log_metric("Training mean_recall", mean_recall, epoch=epoch + 1)

        # cometml_experiemnt.log_metrics({f"Training Recall class {str(x)}": recall[x] for x in range(len(recall))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"Training Precision class {str(x)}": precision[x] for x in range(len(precision))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"Training F1_score class {str(x)}": classwise_f1_score[x] for x in range(len(classwise_f1_score))}, epoch=epoch+1)

        # cometml_experiemnt.log_metrics({f"L1 Training Recall class {str(x)}": l1_recall[x] for x in range(len(l1_recall))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L1 Training Precision class {str(x)}": l1_precision[x] for x in range(len(l1_precision))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L1 Training F1_score class {str(x)}": l1_f1[x] for x in range(len(l1_f1))}, epoch=epoch+1)

        # cometml_experiemnt.log_metrics({f"L2 Training Recall class {str(x)}": l2_recall[x] for x in range(len(l2_recall))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L2 Training Precision class {str(x)}": l2_precision[x] for x in range(len(l2_precision))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L2 Training F1_score class {str(x)}": l2_f1[x] for x in range(len(l2_f1))}, epoch=epoch+1)

        cometml_experiemnt.log_metrics({f"Concat Training Recall class {str(x)}": recall[x] for x in range(len(recall))}, epoch=epoch + 1)
        cometml_experiemnt.log_metrics({f"Concat Training Precision class {str(x)}": precision[x] for x in range(len(precision))}, epoch=epoch + 1)
        cometml_experiemnt.log_metrics({f"Concat Training F1_score class {str(x)}": classwise_f1_score[x] for x in range(len(classwise_f1_score))}, epoch=epoch + 1)

        print("Training Epoch {0:2d} average loss: {1:1.2f}".format(epoch + 1, total_loss / self.train_loader.__len__()))

        return total_loss / self.train_loader.__len__()

    def validate(self, epoch: int, cometml_experiemnt: object, save_individual_plots_specific: bool = False) -> tuple:
        """validating single epoch

        Args:
            epoch (int): current epoch
            cometml_experiemnt (object): logging experiment

        Returns:
            tuple: (validation loss, mIoU)
        """
        print("validating")
        total_loss = 0
        preds, stacked_preds, targets = [], [], []
        histogram_distance, l1_dist, l2_dist = [], [], []
        accuracies = 0
        running_ap = 0.0
        batch_index_to_show = config['visualization']['batch_index_to_show']
        if self.test_with_full_supervision == 1:
            loader = self.test_loader
        else:
            loader = self.val_loader
        loader_size = len(loader)
        if config['visualization']['val_visualization_divisor'] > loader_size:
            config['visualization']['val_visualization_divisor'] = loader_size
        iter_visualization = loader_size // config['visualization']['val_visualization_divisor']
        self.model.eval()
        with torch.no_grad():
            pbar = tqdm(enumerate(loader), total=len(loader))
            for batch_index, batch in pbar:
                outputs = batch
                image, mask = outputs['inputs']['im'].data[0], batch['label']['class_masks'].data[0]
                # print(f"image min:{image.min()}, max:{image.max()}")
                image_name = f"{str(outputs['tr'].data[0][batch_index_to_show]['gids'])}_{str(outputs['tr'].data[0][batch_index_to_show]['slices'])}"
                # original_width, original_height = outputs['tr'].data[0][batch_index_to_show]['space_dims']
                mask = torch.stack(mask)
                mask = mask.long().squeeze(1)
                mask[mask == -1] = 0
                mask[mask > 4] = 0

                bs, c, t, h, w = image.shape

                class_to_show = max(0, torch.unique(mask)[-1] - 1)
                image = image.to(device).squeeze(2)
                mask = mask.to(device)

                # image1 = utils.stad_image(image1)
                # image2 = utils.stad_image(image2)
                image = F.normalize(image, dim=1, p=1)

                output = self.model(image)
                masks = F.softmax(output, dim=1)
                pred1 = masks.max(1)[1].cpu().detach()  # .numpy()
                preds.append(pred1)

                targets.append(mask.cpu())  # .numpy())

                if config['visualization']['val_visualizer'] or (config['visualization']['save_individual_plots'] and save_individual_plots_specific):
                    if (epoch) % config['visualization']['visualize_val_every'] == 0:
                        if (batch_index % iter_visualization) == 0:
                            figure = plt.figure(figsize=(config['visualization']['fig_size'], config['visualization']['fig_size']),
                                                dpi=config['visualization']['dpi'])
                            ax1 = figure.add_subplot(1, 4, 1)
                            ax2 = figure.add_subplot(1, 4, 2)
                            ax3 = figure.add_subplot(1, 4, 3)
                            ax4 = figure.add_subplot(1, 4, 4)
                            # ax5 = figure.add_subplot(3, 3, 5)
                            # ax6 = figure.add_subplot(3, 3, 6)
                            # ax7 = figure.add_subplot(3, 3, 7)
                            # ax8 = figure.add_subplot(3, 3, 8)
                            # ax9 = figure.add_subplot(3, 3, 9)
                            # ax10 = figure.add_subplot(3, 4, 10)
                            # ax11 = figure.add_subplot(3, 4, 11)
                            # ax12 = figure.add_subplot(3, 4, 12)

                            cmap_gradients = plt.cm.get_cmap('jet')
                            # image_show = np.transpose(image1.cpu().detach().numpy()[batch_index_to_show,:,:,:],(1,2,0))[:,:1:4,:3]
                            image_show1 = np.transpose(image.cpu().detach().numpy()[batch_index_to_show, :, :, :], (1, 2, 0))[:, :, :3]
                            image_show1 = np.flip(image_show1, axis=2)

                            image_show1 = (image_show1 - image_show1.min()) / (image_show1.max() - image_show1.min())
                            gt_mask_show1 = mask.cpu().detach()[batch_index_to_show, :, :].numpy().squeeze()

                            # histograms_intersection_show = (histograms_intersection_show - histograms_intersection_show.min())/(histograms_intersection_show.max() - histograms_intersection_show.min())

                            pred1_show = masks.max(1)[1].cpu().detach().numpy()[batch_index_to_show, :, :]
                            pred_fp_tp_fn_prediction_mask = gt_mask_show1 + (2 * pred1_show)

                            classes_in_gt = np.unique(gt_mask_show1)
                            ax1.imshow(image_show1)

                            ax2.imshow(image_show1)
                            ax2.imshow(gt_mask_show1, cmap=self.cmap, vmin=0, vmax=self.max_label)

                            masks_show = masks.cpu().detach().numpy()[batch_index_to_show, 2, :, :]
                            ax3.imshow(masks_show)

                            ax4.imshow(pred1_show, cmap=self.cmap, vmin=0, vmax=self.max_label)

                            # ax1.axis('off')
                            # ax2.axis('off')
                            # ax3.axis('off')
                            # ax4.axis('off')
                            # ax5.axis('off')
                            # ax6.axis('off')
                            # ax7.axis('off')
                            # ax8.axis('off')
                            # ax9.axis('off')

                            if config['visualization']['titles']:
                                ax1.set_title(f"Input Image 1", fontsize=config['visualization']['font_size'])
                                ax2.set_title(f"Input Image 2", fontsize=config['visualization']['font_size'])
                                ax3.set_title(f"Change GT Mask overlaid", fontsize=config['visualization']['font_size'])
                                ax4.set_title(f"l1_patched_diff_change_features_show", fontsize=config['visualization']['font_size'])
                                # ax5.set_title(f"l2_patched_diff_change_features_show", fontsize=config['visualization']['font_size'])
                                # ax6.set_title(f"histograms_intersection_show", fontsize=config['visualization']['font_size'])
                                # ax7.set_title(f"l1_fp_tp_fn_prediction_mask", fontsize=config['visualization']['font_size'])
                                # ax8.set_title(f"l2_fp_tp_fn_prediction_mask", fontsize=config['visualization']['font_size'])
                                # ax9.set_title(f"histc_fp_tp_fn_prediction_mask", fontsize=config['visualization']['font_size'])
                                # figure.suptitle(
                                #     f"Epoch: {epoch+1}\nGT labels for classification: {classes_in_gt}, \nunique in change predictions: {np.unique(change_detection_show)}\nunique in predictions1: {np.unique(logits_show1)}", fontsize=config['visualization']['font_size'])

                            figure.tight_layout()
                            if config['visualization']['val_imshow']:
                                plt.show()

                            if (config['visualization']['save_individual_plots'] and save_individual_plots_specific):

                                plots_path_save = f"{config['visualization']['save_individual_plots_path']}{config['dataset']}/"
                                fig_save_image_root = (f"{plots_path_save}/image_root/", ax1)
                                fig_save_prediction_root = (f"{plots_path_save}/predictions/", ax3)
                                fig_save_overlaid_full_supervised_mask_on_image_alpha_with_bg = (f"{plots_path_save}/overlaid_full_alpha_w_bg/", ax2)
                                roots = [
                                    fig_save_image_root,
                                    fig_save_prediction_root,
                                    fig_save_overlaid_full_supervised_mask_on_image_alpha_with_bg
                                ]
                                figure.savefig(
                                    f"{plots_path_save}/figs/{image_name}.png", bbox_inches='tight')
                                for root, ax in roots:
                                    utils.create_dir_if_doesnt_exist(root)
                                    file_path = f"{root}/{image_name}.png"
                                    # extent = ax.get_window_extent().transformed(figure.dpi_scale_trans.inverted())
                                    extent = ax.get_tightbbox(figure.canvas.get_renderer()).transformed(figure.dpi_scale_trans.inverted())
                                    figure.savefig(file_path, bbox_inches=extent)

                            cometml_experiemnt.log_figure(figure_name=f"Validation, Image name: {image_name}", figure=figure)
                            figure.clear()
                            plt.cla()
                            plt.clf()
                            plt.close('all')
                            plt.close(figure)
                            gc.collect()

        mean_iou, precision, recall = eval_utils.compute_jaccard(preds, targets, num_classes=config['data']['num_classes'])

        # hist_mean_iou, hist_precision, hist_recall = eval_utils.compute_jaccard(histogram_distance, targets, num_classes=2)
        # l1_mean_iou, l1_precision, l1_recall = eval_utils.compute_jaccard(l1_dist, targets, num_classes=2)
        # l2_mean_iou, l2_precision, l2_recall = eval_utils.compute_jaccard(l2_dist, targets, num_classes=2)

        # l1_precision = np.array(l1_precision)
        # l1_recall = np.array(l1_recall)
        # l1_f1 = 2 * (l1_precision * l1_recall) / (l1_precision + l1_recall)

        # l2_precision = np.array(l2_precision)
        # l2_recall = np.array(l2_recall)
        # l2_f1 = 2 * (l2_precision * l2_recall) / (l2_precision + l2_recall)

        # hist_precision = np.array(hist_precision)
        # hist_recall = np.array(hist_recall)
        # hist_f1 = 2 * (hist_precision * hist_recall) / (hist_precision + hist_recall)

        mean_iou = np.array(mean_iou)
        precision = np.array(precision)
        recall = np.array(recall)
        classwise_f1_score = 2 * (precision * recall) / (precision + recall + 0.00000000001)

        mean_precision = precision.mean()
        mean_recall = recall.mean()
        overall_miou = mean_iou.mean()
        mean_f1_score = classwise_f1_score.mean()
        mean_f1_score_from_r_p = 2 * (mean_precision * mean_recall) / (mean_precision + mean_recall + 0.00000000001)

        # print("Validation Epoch {0:2d} average loss: {1:1.2f}".format(epoch+1, total_loss/loader.__len__()))
        # cometml_experiemnt.log_metric("Validation mIoU", overall_miou, epoch=epoch+1)
        cometml_experiemnt.log_metric("Validation mean precision", mean_precision, epoch=epoch + 1)
        cometml_experiemnt.log_metric("Validation mean recall", mean_recall, epoch=epoch + 1)
        cometml_experiemnt.log_metric("Validation mean f1_score", mean_f1_score, epoch=epoch + 1)
        cometml_experiemnt.log_metric("Validation mean f1_score_rp", mean_f1_score_from_r_p, epoch=epoch + 1)

        print({f"Recall class {x}": str(np.float64(recall[x])) for x in range(len(recall))})
        print({f"Precision class {x}": str(np.float64(precision[x])) for x in range(len(precision))})
        print({f"F1 class {x}": str(np.float64(classwise_f1_score[x])) for x in range(len(classwise_f1_score))})

        print(f"Mean F1 score: {mean_f1_score}")
        print(f"Mean F1 score from average recall and precision: {mean_f1_score_from_r_p}")
        print(f"Mean Recall: {mean_recall}")
        print(f"Mean Precision: {mean_precision}")

        # cometml_experiemnt.log_metrics({f"Recall class {str(x)}": recall[x] for x in range(len(recall))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"Precision class {str(x)}": precision[x] for x in range(len(precision))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"F1_score class {str(x)}": classwise_f1_score[x] for x in range(len(classwise_f1_score))}, epoch=epoch+1)

        # cometml_experiemnt.log_metrics({f"L1 Validation Recall class {str(x)}": l1_recall[x] for x in range(len(l1_recall))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L1 Validation Precision class {str(x)}": l1_precision[x] for x in range(len(l1_precision))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L1 Validation F1_score class {str(x)}": l1_f1[x] for x in range(len(l1_f1))}, epoch=epoch+1)

        # cometml_experiemnt.log_metrics({f"L2 Validation Recall class {str(x)}": l2_recall[x] for x in range(len(l2_recall))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L2 Validation Precision class {str(x)}": l2_precision[x] for x in range(len(l2_precision))}, epoch=epoch+1)
        # cometml_experiemnt.log_metrics({f"L2 Validation F1_score class {str(x)}": l2_f1[x] for x in range(len(l2_f1))}, epoch=epoch+1)

        cometml_experiemnt.log_metrics({f"Concat Validation Recall class {str(x)}": recall[x] for x in range(len(recall))}, epoch=epoch + 1)
        cometml_experiemnt.log_metrics({f"Concat Validation Precision class {str(x)}": precision[x] for x in range(len(precision))}, epoch=epoch + 1)
        cometml_experiemnt.log_metrics({f"Concat Validation F1_score class {str(x)}": classwise_f1_score[x] for x in range(len(classwise_f1_score))}, epoch=epoch + 1)

        cometml_experiemnt.log_metric("Validation Average Loss", total_loss / loader.__len__(), epoch=epoch + 1)

        return total_loss / loader.__len__(), classwise_f1_score

    def forward(self, cometml_experiment: object, world_size: int = 8) -> tuple:
        """forward pass for all epochs

        Args:
            cometml_experiment (object): comet ml experiment for logging
            world_size (int, optional): for distributed training. Defaults to 8.

        Returns:
            tuple: (train losses, validation losses, mIoU)
        """
        train_losses, val_losses = [], []
        mean_ious_val, mean_ious_val_list, count_metrics_list = [], [], []
        best_val_loss, best_train_loss, train_loss = np.infty, np.infty, np.infty
        best_val_mean_f1, val_mean_f1 = 0, 0
        best_val_change_f1, val_change_f1 = 0, 0

        model_save_dir = config['data'][config['location']]['model_save_dir'] + \
            f"{current_path[-1]}_{config['dataset']}/{cometml_experiment.project_name}_{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M')}/"
        utils.create_dir_if_doesnt_exist(model_save_dir)
        for epoch in range(0, self.epochs):
            if config['procedures']['train']:
                with cometml_experiment.train():
                    train_loss = self.train(epoch, cometml_experiment)
            if config['procedures']['validate']:
                with cometml_experiment.validate():
                    val_loss, val_cw_f1 = self.validate(
                        epoch, cometml_experiment)
                    val_mean_f1 = val_cw_f1.mean()
                    val_change_f1 = val_cw_f1[1]
            self.scheduler.step()

            if (train_loss <= best_train_loss) or (val_mean_f1 >= best_val_mean_f1) or (val_change_f1 >= best_val_change_f1):

                if train_loss <= best_train_loss:
                    best_train_loss = train_loss
                if val_mean_f1 >= best_val_mean_f1:
                    best_val_mean_f1 = val_mean_f1
                if val_change_f1 >= best_val_change_f1:
                    best_val_change_f1 = val_change_f1

                model_save_name = f"{current_path[-1]}_epoch_{epoch}_loss_{train_loss}_valmF1_{val_mean_f1}_valChangeF1_{val_change_f1}_time_{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')}.pth"

                if config['procedures']['train']:
                    config['val_change_f1'] = str(val_change_f1)
                    config['val_mean_f1'] = str(val_mean_f1)
                    config['train_loss'] = str(train_loss)
                    with open(model_save_dir + "config.yaml", 'w') as file:
                        yaml.dump(config, file)

                    torch.save({'epoch': epoch,
                                'model': self.model.state_dict(),
                                'optimizer': self.optimizer.state_dict(),
                                'scheduler': self.scheduler.state_dict(),
                                'loss': train_loss},
                               model_save_dir + model_save_name)
                if config['visualization']['save_individual_plots']:
                    _, _ = self.validate(
                        epoch, cometml_experiment, save_individual_plots_specific=True)

        return train_losses, val_losses, mean_ious_val


if __name__ == "__main__":
    project_root = '/'.join(current_path[:-1])
    # main_config_path = f"{os.getcwd()}/configs/main.yaml"
    main_config_path = f"{project_root}/configs/main.yaml"

    initial_config = utils.load_yaml_as_dict(main_config_path)
    # experiment_config_path = f"{os.getcwd()}/configs/{initial_config['dataset']}.yaml"
    experiment_config_path = f"{project_root}/configs/{initial_config['dataset']}.yaml"
    # config_path = utils.dictionary_contents(os.getcwd()+"/",types=["*.yaml"])[0]

    experiment_config = utils.config_parser(
        experiment_config_path, experiment_type="training")
    config = {**initial_config, **experiment_config}
    config['start_time'] = datetime.datetime.today().strftime(
        '%Y-%m-%d-%H:%M:%S')

    # _{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M')}"
    project_name = f"{current_path[-3]}_{current_path[-1]}_{config['dataset']}"
    experiment_name = f"attention_{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')}"
    experiment = comet_ml.Experiment(api_key=config['cometml']['api_key'],
                                     project_name=project_name,
                                     workspace=config['cometml']['workspace'],
                                     display_summary_level=0)

    config['experiment_url'] = str(experiment.url)

    experiment.set_name(experiment_name)

    torch.manual_seed(config['seed'])
    torch.cuda.manual_seed(config['seed'])
    np.random.seed(config['seed'])
    random.seed(config['seed'])
    torch.set_default_dtype(torch.float32)

    device_ids = list(range(torch.cuda.device_count()))
    # print(device_ids)
    # config['device_ids'] = device_ids
    # gpu_devices = ','.join([str(id) for id in device_ids])
    # os.environ["CUDA_VISIBLE_DEVICES"] = gpu_devices
    device = torch.device('cuda')

    # config['devices_used'] = gpu_devices
    experiment.log_asset_data(config)
    experiment.log_text(config)
    experiment.log_parameters(config)
    experiment.log_parameters(config['training'])
    experiment.log_parameters(config['evaluation'])
    experiment.log_parameters(config['visualization'])

    if config['visualization']['train_imshow'] or config['visualization']['val_imshow']:
        matplotlib.use('TkAgg')

    if config['training']['resume'] != False:
        base_path = '/'.join(config['training']['resume'].split('/')[:-1])
        pretrain_config_path = f"{base_path}/config.yaml"
        pretrain_config = utils.load_yaml_as_dict(pretrain_config_path)
        # print(config['training']['model_feats_channels'])
        # print(pretrain_config_path['training']['model_feats_channels'])
        config['data']['channels'] = pretrain_config['data']['channels']
        # if not config['training']['model_feats_channels'] == pretrain_config_path['training']['model_feats_channels']:
        #     print("the loaded model does not have the same number of features as configured in the experiment yaml file. Matching channel sizes to the loaded model instead.")
        # config['training']['model_feats_channels'] = pretrain_config_path['training']['model_feats_channels']
        config['data']['num_classes'] = pretrain_config['data']['num_classes']
        config['training']['model_feats_channels'] = pretrain_config['training']['model_feats_channels']

    if config['data']['name'] == 'watch' or config['data']['name'] == 'onera':
        coco_fpath = ub.expandpath(config['data'][config['location']]['train_coco_json'])
        dset = kwcoco.CocoDataset(coco_fpath)
        print(dset.cats)
        sampler = ndsampler.CocoSampler(dset)

        window_dims = (config['data']['time_steps'], config['data']['image_size'], config['data']['image_size'])  # [t,h,w]
        input_dims = (config['data']['image_size'], config['data']['image_size'])

        channels = config['data']['channels']
        num_channels = len(channels.split('|'))
        config['training']['num_channels'] = num_channels

        dataset = SequenceDataset(sampler, window_dims, input_dims, channels)
        train_dataloader = dataset.make_loader(batch_size=config['training']['batch_size'])

        test_coco_fpath = ub.expandpath(config['data'][config['location']]['test_coco_json'])
        test_dset = kwcoco.CocoDataset(test_coco_fpath)
        test_sampler = ndsampler.CocoSampler(test_dset)

        test_dataset = SequenceDataset(test_sampler, window_dims, input_dims, channels)
        test_dataloader = test_dataset.make_loader(batch_size=config['evaluation']['batch_size'])
    else:
        train_dataloader = build_dataset(dataset_name=config['data']['name'],
                                        root=config['data'][config['location']]['train_dir'],
                                        batch_size=config['training']['batch_size'],
                                        num_workers=config['training']['num_workers'],
                                        split='train',
                                        crop_size=config['data']['image_size'],
                                        channels=config['data']['channels'],
                                        )

        channels = config['data']['channels']
        num_channels = len(channels.split('|'))
        config['training']['num_channels'] = num_channels
        window_dims = (config['data']['time_steps'], config['data']['image_size'], config['data']['image_size'])  # [t,h,w]
        input_dims = (config['data']['image_size'], config['data']['image_size'])

        test_coco_fpath = ub.expandpath(config['data'][config['location']]['test_coco_json'])
        test_dset = kwcoco.CocoDataset(test_coco_fpath)
        test_sampler = ndsampler.CocoSampler(test_dset)

        test_dataset = SequenceDataset(test_sampler, window_dims, input_dims, channels)
        test_dataloader = test_dataset.make_loader(batch_size=config['evaluation']['batch_size'])

    # if not config['training']['model_single_input'] and not config['training']['model_diff_input']:
    #     config['training']['num_channels'] = 2*config['training']['num_channels']

    model = build_model(model_name=config['training']['model_name'],
                        backbone=config['training']['backbone'],
                        pretrained=config['training']['pretrained'],
                        num_classes=config['data']['num_classes'],
                        num_groups=config['training']['gn_n_groups'],
                        weight_std=config['training']['weight_std'],
                        beta=config['training']['beta'],
                        num_channels=config['training']['num_channels'],
                        out_dim=config['training']['out_features_dim'],
                        feats=config['training']['model_feats_channels'])

    # model = SupConResNet(name=config['training']['backbone'])
    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("model has {} trainable parameters".format(num_params))
    model = nn.DataParallel(model)
    model.to(device)

    optimizer = optim.SGD(model.parameters(),
                          lr=config['training']['learning_rate'],
                          momentum=config['training']['momentum'],
                          weight_decay=config['training']['weight_decay'])

    scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer, len(train_dataloader),
                                                     eta_min=config['training']['learning_rate'])

    if config['training']['resume'] != False:

        if os.path.isfile(config['training']['resume']):
            checkpoint = torch.load(config['training']['resume'])
            # model_dict = model.state_dict()
            # if model_dict == checkpoint['model']:
            #     print(f"Succesfuly loaded model from {config['training']['resume']}")
            # else:
            #     pretrained_dict = {k: v for k, v in checkpoint['model'].items() if k in model_dict}
            #     model_dict.update(pretrained_dict)
            #     model.load_state_dict(model_dict)
            #     print("There was model mismatch. Matching elements in the pretrained model were loaded.")
            missing_keys, unexpexted_keys = model.load_state_dict(checkpoint['model'], strict=False)
            start_epoch = checkpoint['epoch']
            optimizer.load_state_dict(checkpoint['optimizer'])
            scheduler.load_state_dict(checkpoint['scheduler'])
            print(f"loadded model succeffuly from: {config['training']['resume']}")
            print(f"Missing keys from loaded model: {missing_keys}, unexpected keys: {unexpexted_keys}")
        else:
            print("no checkpoint found at {}".format(
                config['training']['resume']))
            exit()

    trainer = Trainer(model,
                      train_dataloader,
                      test_dataloader,
                      config['training']['epochs'],
                      optimizer,
                      scheduler,
                      test_loader=test_dataloader,
                      test_with_full_supervision=config['training']['test_with_full_supervision']
                      )
    train_losses, val_losses, mean_ious_val = trainer.forward(experiment)



import string


def partial_format(format_string, *args, **kwargs):
    """
    A solution to the partial string formatting problem.

    Taken from [SO11283961]_, which is a modification of the stdlib
    string.Formatter code.

    Args:
        format_string (str): the templated string to be formatted
        *args: positional replacements
        **kwargs: key value replacements

    Returns:
        str : the format string with only the specified parts replaced.

    Example:
        >>> from watch.utils.partial_format import partial_format
        >>> format_string = '{foo} + {bar} = {baz}'
        >>> args = tuple()
        >>> kwargs = dict(bar=3)
        >>> partial_format(format_string, *args, **kwargs)
        '{foo} + 3 = {baz}'

    References:
        [SO11283961] https://stackoverflow.com/questions/11283961/partial-string-formatting
    """
    return _PartialFormatter().format(format_string, *args, **kwargs)


class _PartialFormatter(string.Formatter):
    """
    A modified string formatter that handles a partial set of format
    args/kwargs.
    """

    def vformat(self, format_string, args, kwargs):
        used_args = set()
        result, _ = self._vformat(format_string, args, kwargs, used_args, 2)
        self.check_unused_args(used_args, args, kwargs)
        return result

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth,
                 auto_arg_index=0):
        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')
        result = []
        for literal_text, field_name, format_spec, conversion in \
                self.parse(format_string):

            orig_field_name = field_name

            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do
                #  the formatting

                # handle arg indexing when empty field_names are given.
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # given the field_name, find the object it references
                #  and the argument it came from
                try:
                    obj, arg_used = self.get_field(field_name, args, kwargs)
                except (IndexError, KeyError):
                    ##########################
                    # Where the magic happens.
                    # ------------------------
                    # This case is the main difference between this class and
                    # the stdlib implementation.
                    ##########################
                    # catch issues with both arg indexing and kwarg key errors
                    obj = orig_field_name
                    if conversion:
                        obj += '!{}'.format(conversion)
                    if format_spec:
                        format_spec, auto_arg_index = self._vformat(
                            format_spec, args, kwargs, used_args,
                            recursion_depth, auto_arg_index=auto_arg_index)
                        obj += ':{}'.format(format_spec)
                    result.append('{' + obj + '}')
                else:
                    used_args.add(arg_used)

                    # do any conversion on the resulting object
                    obj = self.convert_field(obj, conversion)

                    # expand the format spec, if needed
                    format_spec, auto_arg_index = self._vformat(
                        format_spec, args, kwargs,
                        used_args, recursion_depth - 1,
                        auto_arg_index=auto_arg_index)

                    # format the object and append to the result
                    result.append(self.format_field(obj, format_spec))

        return ''.join(result), auto_arg_index


def _test_partial_format():
    """
    Example:
        _test_partial_format
    """
    import pytest

    def test_auto_indexing():
        # test basic arg auto-indexing
        assert partial_format('{}{}', 4, 2) == '42'
        assert partial_format('{}{} {}', 4, 2) == '42 {}'

    def test_manual_indexing():
        # test basic arg indexing
        assert partial_format('{0}{1} is not {1} or {0}', 4, 2) == '42 is not 2 or 4'
        assert partial_format('{0}{1} is {3} {1} or {0}', 4, 2) == '42 is {3} 2 or 4'

    def test_mixing_manualauto_fails():
        # test mixing manual and auto args raises
        with pytest.raises(ValueError):
            assert partial_format('{!r} is {0}{1}', 4, 2)

    def test_kwargs():
        # test basic kwarg
        assert partial_format('{base}{n}', base=4, n=2) == '42'
        assert partial_format('{base}{n}', base=4, n=2, extra='foo') == '42'
        assert partial_format('{base}{n} {key}', base=4, n=2) == '42 {key}'

    def test_args_and_kwargs():
        # test mixing args/kwargs with leftovers
        assert partial_format('{}{k} {v}', 4, k=2) == '42 {v}'

        # test mixing with leftovers
        r = partial_format('{}{} is the {k} to {!r}', 4, 2, k='answer')
        assert r == '42 is the answer to {!r}'

    def test_coercion():
        # test coercion is preserved for skipped elements
        assert partial_format('{!r} {k!r}', '42') == "'42' {k!r}"

    def test_nesting():
        # test nesting works with or with out parent keys
        assert partial_format('{k:>{size}}', k=42, size=3) == ' 42'
        assert partial_format('{k:>{size}}', size=3) == '{k:>3}'

    test_mixing_manualauto_fails()
    test_auto_indexing()
    test_manual_indexing()
    test_kwargs()
    test_args_and_kwargs()
    test_coercion()
    test_nesting()

    cases = [
        ('{a} {b}', '1 2.0'),
        ('{z} {y}', '{z} {y}'),
        ('{a} {a:2d} {a:04d} {y:2d} {z:04d}', '1  1 0001 {y:2d} {z:04d}'),
        ('{a!s} {z!s} {d!r}', '1 {z!s} {\'k\': \'v\'}'),
        ('{a!s:>2s} {z!s:>2s}', ' 1 {z!s:>2s}'),
        ('{a!s:>{a}s} {z!s:>{z}s}', '1 {z!s:>{z}s}'),
        ('{a.imag} {z.y}', '0 {z.y}'),
        ('{e[0]:03d} {z[0]:03d}', '042 {z[0]:03d}'),
    ]

    for s, expected in cases:
        # test a bunch of random stuff
        data = dict(
            a=1,
            b=2.0,
            c='3',
            d={'k': 'v'},
            e=[42],
        )
        result = partial_format(s, **data)
        assert expected == result


def subtemplate(text, subs=None, /, **kwargs):
    """
    Substitutes variables into a templated string.

    Similar to format, replaces variables with bash-like dollar sign patterns
    (e.g. $VARNAME or ${VARNAME}).

    Args:
        text (str): the templated text
        subs (dict | None): a dictionary of substitutions
        **kwargs: other substitutions

    Returns:
        str: formatted text. Unspecified variables are left unchanged.

    Notes:
        Order of precidence from lowest to highest goes: subs, kwargs

    Example:
        >>> from watch.utils.partial_format import *  # NOQA
        >>> import ubelt as ub
        >>> text = ub.codeblock(
        >>>     '''
        >>>     The $SUBJECT $VERB the $OBJECT
        >>>     ''')
        >>> print(subtemplate(text, SUBJECT='dog', OBJECT='food'))
        The dog $VERB the food
        >>> print(subtemplate(text, SUBJECT='dog', VERB='eats', OBJECT='food'))
        The dog eats the food
        >>> print(subtemplate(text, {'SUBJECT': 'dude'}, SUBJECT='dog'))
        The dog $VERB the $OBJECT
    """
    import string
    import ubelt as ub
    import operator as op
    from functools import reduce
    fmtdict = ub.udict({})
    if subs:
        fmtdict.update(subs)
    fmtdict.update(kwargs)
    if not fmtdict:
        return text
    template = string.Template(text)
    existing_vars = {reduce(op.add, t, '') for t in template.pattern.findall(text)}
    fmtdict = fmtdict & existing_vars
    return template.safe_substitute(**fmtdict)


def fsubtemplate(text, subs=None, /, **kwargs):
    """
    Like subtemplate, but uses the current local variable context

    Args:
        text (str): the templated text
        subs (dict | None): a dictionary of substitutions
        **kwargs: other substitutions

    Returns:
        str: formatted text. Unspecified variables are left unchanged.

    Notes:
        Order of precidence from lowest to highest goes: locals, subs, kwargs

    Example:
        >>> from watch.utils.partial_format import *  # NOQA
        >>> import ubelt as ub
        >>> text = ub.codeblock(
        >>>     '''
        >>>     The $SUBJECT $VERB the $OBJECT
        >>>     ''')
        >>> SUBJECT = 'dude'
        >>> print(fsubtemplate(text, OBJECT='food'))
        The dude $VERB the food
        >>> print(fsubtemplate(text, SUBJECT='dog', OBJECT='food'))
        The dog $VERB the food
        >>> print(fsubtemplate(text, SUBJECT='dog', VERB='eats', OBJECT='food'))
        The dog eats the food
    """
    import ubelt as ub
    locals_ = _get_stack_frame(1).f_locals
    subs = ub.udict()
    subs.update(locals_)
    if subs is not None:
        subs.update(subs)
    subs.update(kwargs)
    return subtemplate(text, subs)


def _get_stack_frame(N=0, strict=True):
    """
    Args:
        N (int): N=0 means the frame you called this function in.
                 N=1 is the parent frame.
        strict (bool): (default = True)
    """
    import inspect
    frame_cur = inspect.currentframe()
    for idx in range(N + 1):
        # always skip the frame of this function
        frame_next = frame_cur.f_back
        if frame_next is None:
            if strict:
                raise AssertionError('Frame level {:!r} is root'.format(idx))
            else:
                break
        frame_cur = frame_next
    return frame_cur



# FIXME:
# Adds the "modules" subdirectory to the python path.
# See https://gitlab.kitware.com/smart/watch/-/merge_requests/148#note_1050127
# for discussion of how to refactor this in the future.
from . import modules_monkeypatch  # NOQA

import warnings
import numpy as np
import torch
import torch.nn as nn
import torchvision.transforms

import pytorch_lightning as pl

from .backbone import get_backbone

from frame_field_learning import data_transforms
from frame_field_learning import local_utils
from frame_field_learning.model_multi import Multi_FrameFieldModel

from medpy.filter.smoothing import anisotropic_diffusion
from scipy import ndimage

dfactor = 25.5

# device = torch.device("cuda" if torch.cuda.is_available() else "cpu") # PyTorch v0.4.0


#-------------------------------------------
# Modify the batch_norm layers
#-------------------------------------------

def modify_bn(model, track_running_stats=True, bn_momentum=0.1):
    for m in model.modules():
        for child in m.children():
            if type(child) == nn.BatchNorm2d:

                child.momentum = bn_momentum
                child.track_running_stats = track_running_stats

                if track_running_stats is False:
                    child.running_mean = None
                    child.running_var =  None

    return model


#-------------------------------------------------
# Depth/Label/Shadow/Facade Eestimation Module
#-------------------------------------------------

class MultiTaskModel(pl.LightningModule):

    def __init__(
        self,
        batch_size: int = 1,
        checkpoint: str = None,
        config: dict = None,
        test_img_dir: str = None,
        test_img_list: str = None,
        gpus: str = '0',
        **kwargs,
    ):

        super().__init__(**kwargs)
        self.gpus = gpus
        self.checkpoint = checkpoint
        self.batch_size = batch_size

        self.test_img_dir = test_img_dir
        self.test_img_list = test_img_list

        self.config = config

        self.backbone = get_backbone(self.config["backbone_params"])

        train_online_cuda_transform = None
        eval_online_cuda_transform = None

        self.net = Multi_FrameFieldModel(
            self.config,
            backbone=self.backbone,
            train_transform=train_online_cuda_transform,
            eval_transform=eval_online_cuda_transform)

        self.transform =  data_transforms.get_online_cuda_transform(
            self.config,
            augmentations=self.config["data_aug_params"]["enable"])

    def forward(self, x, tta=False):
        return self.net(x, tta)

    def test_step(self, batch, batch_idx):

        out_arr = []
        for i, image in enumerate(batch):
            if isinstance(image, dict):
                gid = image['id']
                # img_info = image
                image = image['imgdata']

            with torch.no_grad():

                image_float = image / 255.0
                mean = np.mean(image_float.reshape(-1, image_float.shape[-1]), axis=0)
                std = np.std(image_float.reshape(-1, image_float.shape[-1]), axis=0)

                batch2 = {
                    "image": torchvision.transforms.functional.to_tensor(image)[None, ...],
                    "image_mean": torch.from_numpy(mean)[None, ...],
                    "image_std": torch.from_numpy(std)[None, ...],
                }

                batch2 = local_utils.batch_to_cuda(batch2)

                pred2, batch2 = self(batch2, tta=True)

                output_depth = pred2['depth'][0, 0, :, :].cpu().data.numpy()
                output_label = pred2['seg'][0, 0, :, :].cpu().data.numpy()

                weighted_depth = dfactor * output_depth

                alpha = 0.9
                weighted_seg = alpha * output_label + (1.0 - alpha) * np.minimum(0.99, weighted_depth / 70.0)

                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    tmp2 = 255 * anisotropic_diffusion(weighted_seg, niter=1, kappa=100, gamma=0.8)
                weighted_final = ndimage.median_filter(tmp2.astype(np.uint8), size=7)

                # Image.fromarray(weighted_final.astype(np.uint8)).save('/output/weighted_final.png')

            out_arr.append((gid, weighted_final))
        return out_arr

    @staticmethod
    def add_model_specific_args(parent_parser):  # pragma: no-cover
        parser = parent_parser.add_argument_group("MultiTaskModel")

        parser.add_argument('--checkpoint', default=None, type=str,
                            help='checkpoint to use for testing')
        parser.add_argument('--config', '--config', default=None, type=str,
                            help='Name of the config file, excluding the .json file extension.')
        parser.add_argument('--test_img_dir', '--test_img_dir', default=None, type=str,
                            help='directory where test images are located')
        parser.add_argument('--test_img_list', '--test_img_list', default=None, type=str,
                            help='list of test images')
        parser.add_argument('--gpus', default='0', type=str,
                            help='GPU')

        return parent_parser



# flake8: noqa
# from material_seg.datasets import build_dataset
import kwcoco
import ndsampler
import ubelt as ub
# import watch
import numpy as np
import matplotlib.pyplot as plt
# from tqdm import tqdm
import torch
# import pdb
from sklearn.cluster import KMeans
from sklearn.manifold import TSNE
from metric_learn import NCA, LMNN, MMC_Supervised, LSML_Supervised
from watch.tasks.rutgers_material_seg.datasets.iarpa_dataset import SequenceDataset

visualize_images = False
coco_fpath = ub.expandpath('/home/native/core534_data/datasets/smart_watch/processed/drop0_aligned_v2/material_labels.kwcoco.json')
dset = kwcoco.CocoDataset(coco_fpath)

sampler = ndsampler.CocoSampler(dset)

# # print(sampler)
number_of_timestamps, h, w = 4, 128, 128
window_dims = (number_of_timestamps, h, w)  # [t,h,w]
input_dims = (h, w)

# # channels = 'r|g|b|gray|wv1'
channels = 'r|g|b'
# channels = 'gray'
dataset = SequenceDataset(sampler, window_dims, input_dims, channels)
print(dataset.__len__())
loader = dataset.make_loader(batch_size=1)

k = 40
kmeans = KMeans(n_clusters=k, random_state=0, verbose=True)
nca = NCA(max_iter=10, verbose=True, random_state=0)
lmnn = LMNN(k=3, verbose=True, random_state=0, learn_rate=0.01, max_iter=1000)
mmc = MMC_Supervised(max_iter=100, max_proj=10000, verbose=True, random_state=0)
lsml = LSML_Supervised(verbose=True, random_state=0)
n_points = 300

for batch in loader:
    # pdb.set_trace()
    image_data = batch['inputs']['im'].data[0]  # [b,c,t,h,w]
    b, c, t, h, w = image_data.shape
    mask_data = batch['label']['class_masks'].data[0]  # len(mask_data) = b
    mask_data = torch.stack(mask_data)  # .numpy()

    image_show = np.array(image_data).transpose(0, 2, 3, 4, 1) / 500  # visualize 0 indexed in batch
    # plt.imshow(image_show)
    # plt.show()
    # image_show = image_show[0,]
    # mask_show = np.array(mask_data) # [b,t,h,w]

    image_data = image_data.view(b, c * t, h * w)
    mask_data = mask_data.view(b, t, h * w).squeeze(0)
    image_data = torch.transpose(image_data, 1, 2)
    image_data = torch.flatten(image_data, start_dim=0, end_dim=1)

    mask_data = mask_data[0, :]
    # mask_data = mask_data.reshape(h,w)

    # mask_data = mask_data.view(-1,1)
    print(torch.unique(mask_data))
    print(mask_data.shape)
    print(image_data.shape)

    non_bg_indices = np.where(mask_data > 0)[0]
    concrete_indices = np.where(mask_data == 1)[0][:n_points]
    veg_indices = np.where(mask_data == 2)[0][:n_points]
    soil_indices = np.where(mask_data == 3)[0][:n_points]
    water_indices = np.where(mask_data == 4)[0][:n_points]
    material_indices = np.concatenate((concrete_indices, veg_indices, soil_indices, water_indices), axis=0)

    # print(f"concrete: {concrete_indices.shape}, veg: {veg_indices.shape}, soil: {soil_indices.shape}, water: {water_indices.shape}")
    # print(f"concrete: {concrete_indices}, veg: {veg_indices}, soil: {soil_indices}, water: {water_indices}")
    # print(f"concrete: {type(concrete_indices)}, veg: {type(veg_indices)}, soil: {type(soil_indices)}, water: {type(water_indices)}")
    mask_data_filtered = mask_data[material_indices]
    image_data_filtered = image_data[material_indices]

    # nca.fit(image_data[non_bg_indices,:], mask_data[non_bg_indices])
    lmnn.fit(image_data_filtered, mask_data_filtered)
    # mmc.fit(image_data[non_bg_indices,:], mask_data[non_bg_indices])
    # lsml.fit(image_data[non_bg_indices,:], mask_data[non_bg_indices])

    # X_nca = nca.transform(image_data)
    X_lmnn = lmnn.transform(image_data)[material_indices]
    # X_mmc = mmc.transform(image_data)
    # X_lsml = lsml.transform(image_data)
    # print(np.unique(X_lsml))
    # X_lsml_show = X_lsml.reshape(t,h,w,c)/255
    print(X_lmnn.shape)
    # plt.imshow(X_lsml_show[0,:,:,:])
    # plt.show()

    out_feat_embed = TSNE(n_components=2, verbose=True, random_state=0).fit_transform(image_data_filtered)
    # out_feat_embed_nca = TSNE(n_components=2, verbose=True).fit_transform(X_nca)
    out_feat_embed_lmnn = TSNE(n_components=2, verbose=True, random_state=0).fit_transform(X_lmnn)
    # out_feat_embed_mmc = TSNE(n_components=2, verbose=True).fit_transform(X_mmc)
    # out_feat_embed_lsml = TSNE(n_components=2, verbose=True).fit_transform(X_lsml)

    # kmeans.fit(image_data)
    # cluster_centers = kmeans.cluster_centers_
    # cluster_labels = kmeans.labels_
    # y_kmeans = kmeans.predict(image_data)
    # prediction = cluster_labels.reshape(h,w)
    # prediction_no_bg = np.ma.masked_where(prediction==0,prediction)

    fig = plt.figure()
    ax1 = fig.add_subplot(1, 2, 1)
    ax2 = fig.add_subplot(1, 2, 2)
    # ax3 = fig.add_subplot(2,2,3)
    # ax4 = fig.add_subplot(2,2,4)
    scatter1 = ax1.scatter(out_feat_embed[:, 0], out_feat_embed[:, 1], c=mask_data_filtered, marker='.', cmap='Set1')
    legend1 = ax1.legend(*scatter1.legend_elements(), loc="lower left", title="Classes")
    ax1.set_title("TSNE of Image Raw Data")
    # ax2.scatter(out_feat_embed_mmc[:,0], out_feat_embed_mmc[:,1], c=mask_data, marker='.', cmap='Set1')
    # ax3.scatter(out_feat_embed_lsml[:,0], out_feat_embed_lsml[:,1], c=mask_data, marker='.', cmap='Set1')
    scatter2 = ax2.scatter(out_feat_embed_lmnn[:, 0], out_feat_embed_lmnn[:, 1], c=mask_data_filtered, marker='.', cmap='Set1')
    legend2 = ax2.legend(*scatter2.legend_elements(), loc="lower left", title="Classes")
    ax2.set_title("TSNE of Clustered Features")
    plt.show()

    # figure = plt.figure(figsize=(15,15))
    # ax1 = figure.add_subplot(2,5,1)
    # ax2 = figure.add_subplot(2,5,2)
    # ax3 = figure.add_subplot(2,5,3)
    # ax4 = figure.add_subplot(2,5,4)
    # ax5 = figure.add_subplot(2,5,5)
    # ax6 = figure.add_subplot(2,5,6)
    # ax7 = figure.add_subplot(2,5,7)
    # ax8 = figure.add_subplot(2,5,8)
    # ax9 = figure.add_subplot(2,5,9)
    # # ax10 = figure.add_subplot(2,5,10)

    # ax1.imshow(image_show[0,0,:,:,:])
    # ax2.imshow(image_show[0,1,:,:,:])
    # ax3.imshow(image_show[0,2,:,:,:])
    # ax4.imshow(image_show[0,3,:,:,:])
    # # ax1.imshow(image_show[0,0,:,:,:])
    # ax5.imshow(prediction, vmin=0, vmax=k, cmap='Set1')

    # ax6.imshow(mask_data[0,0,:,:])
    # ax7.imshow(mask_data[0,1,:,:])
    # ax8.imshow(mask_data[0,2,:,:])
    # ax9.imshow(mask_data[0,3,:,:])
#     ax4.scatter(x_clusters_scatters, y_clusters_scatters, color=(len(x_clusters_scatters)//channels)*['red','green','blue','yellow','black'])

    plt.show()

    if visualize_images:
        raise NotImplementedError
        mask_show = None
        mask_show = mask_show[0]  # [b,t,h,w]
        image_show = image_show[0]
        figure = plt.figure(figsize=(10, 10))
        axes = {}
        for i in range(1, 2 * t + 1):
            axes[i] = figure.add_subplot(2, t, i)
        for key in axes.keys():
            if key <= t:
                axes[key].imshow(image_show[key - 1, :, :, :])
            else:
                axes[key].imshow(mask_show[key - t - 1, :, :], vmin=-1, vmax=7)
        figure.tight_layout()
        plt.show()



import ubelt as ub
import kwimage
import numpy as np
import ndsampler
import einops
import kwcoco


def demo_visualize_tokenization():
    """
    Helper for slides
    """
    import watch
    from watch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    dvc_dpath = watch.find_smart_dvc_dpath()
    coco_fpath = dvc_dpath / 'Drop1-Aligned-L1-2022-01/combo_DILM_train.kwcoco.json'
    coco_dset = kwcoco.CocoDataset(coco_fpath)

    sampler = ndsampler.CocoSampler(coco_dset)
    channels = 'red|green|blue,nir,bare_ground,panchromatic'
    self = KWCocoVideoDataset(sampler, time_dims=2, window_dims=(96, 96),
                              channels=channels,
                              time_sampling='soft2+distribute',
                              resample_invalid_frames=False)
    self.requested_tasks['change'] = False

    vidid = coco_dset.index.name_to_video['NZ_R001']['id']
    # vidid = list(coco_dset.videos())[3]
    coco_dset.index.videos[vidid]
    images = coco_dset.images(vidid=vidid)
    sensor_gids = {}
    for coco_img in images.coco_images:
        sensor = coco_img.img.get('sensor_coarse')
        if sensor_gids.get(sensor) is None:
            sensor_gids[sensor] = coco_img.img['id']

    gids = list(sensor_gids.values())
    print('gids = {!r}'.format(gids))

    tr = {
        'main_idx': 0,
        'video_id': vidid,
        # 'space_slice': (slice(0, 96, None), slice(0, 96, None)),
        'space_slice': (slice(200, 296, None), slice(200, 296, None)),
        'gids': gids,
    }
    self.disable_augmenter = True
    item = self[tr]

    canvas = self.draw_item(item)
    import kwplot
    kwplot.autompl()
    kwplot.imshow(canvas, fnum=1)

    to_stack_modes = []
    to_stack_flats = []

    for frame in item['frames']:

        frame_stack_modes = []
        frame_stack_flats = []

        for mode_key, mode in frame['modes'].items():
            imtensor = mode.numpy()
            img = einops.rearrange(imtensor, 'c h w -> h w c')
            img = kwimage.normalize_intensity(img)

            chunks1 = einops.rearrange(img, '(wh h) (ww w) c -> (wh ww) h w c', ww=8, wh=8)

            grid_stack = kwimage.stack_images_grid(chunks1, pad=4, axis=0)
            grid_stack = kwimage.imresize(grid_stack, scale=2.0, interpolation='nearest')

            # chunks1 = einops.rearrange(img, '(wh h) (ww w) c -> (wh ww) (c h) w 1', ww=8, wh=8)
            # chunks1 = einops.rearrange(img, '(wh h) (ww w) c -> (wh ww) (h w) 1 c', ww=8, wh=8)
            chunks2 = einops.rearrange(img, '(wh h) (ww f w) c -> (wh ww) (h w) f c', ww=8, wh=8, f=4)

            flat_stack = kwimage.stack_images(chunks2, pad=4, axis=1)

            flat_stack = kwimage.ensure_uint255(flat_stack)
            flat_stack = kwimage.draw_header_text(flat_stack, f'{mode_key}', fit='shrink')

            grid_stack = kwimage.ensure_uint255(grid_stack)
            grid_stack = kwimage.draw_header_text(grid_stack, f'{mode_key}', fit='shrink')

            frame_stack_modes.append(grid_stack)
            frame_stack_flats.append(flat_stack)

        frame_mode_canvas = kwimage.stack_images(frame_stack_modes, axis=1, pad=16, bg_value=(0, 0, 0))
        frame_flat_canvas = kwimage.stack_images(frame_stack_flats, pad=16, axis=1, bg_value=(0, 0, 0))

        sensor = frame['sensor']
        time_index = frame['time_index']
        time_offset = np.array(frame['time_offset']).ravel()[0]

        frame_mode_canvas = kwimage.draw_header_text(frame_mode_canvas, f'{time_index=}', fit='shrink')
        frame_mode_canvas = kwimage.draw_header_text(frame_mode_canvas, f'{time_offset=}', fit='shrink')
        frame_mode_canvas = kwimage.draw_header_text(frame_mode_canvas, f'{sensor=}', fit='shrink')

        frame_flat_canvas = kwimage.draw_header_text(frame_flat_canvas, f'{time_index=}', fit='shrink')
        frame_flat_canvas = kwimage.draw_header_text(frame_flat_canvas, f'{time_offset=}', fit='shrink')
        frame_flat_canvas = kwimage.draw_header_text(frame_flat_canvas, f'{sensor=}', fit='shrink')

        to_stack_modes.append(frame_mode_canvas)
        to_stack_flats.append(frame_flat_canvas)

    canvas_modes = kwimage.stack_images(to_stack_modes, axis=1, pad=16, bg_value=(255, 255, 255))
    canvas_flats = kwimage.stack_images(to_stack_flats, pad=16, axis=1, bg_value=(255, 255, 255))

    kwplot.imshow(canvas_modes, fnum=2)
    kwplot.imshow(canvas_flats, fnum=3)


def demo_visualize_heterogeneous_inputs():
    # xdoctest: +REQUIRES(env:DVC_DPATH)
    # Run the following tests on real watch data if DVC is available
    import watch
    import ubelt as ub
    from watch.utils import kwcoco_extensions
    from watch.tasks.fusion.datamodules.kwcoco_datamodule import KWCocoVideoDataModule
    import kwcoco
    import kwplot
    kwplot.autompl()

    dvc_dpath = watch.find_smart_dvc_dpath()
    coco_fpath = dvc_dpath / 'Drop1-Aligned-L1-2022-01/combo_DILM.kwcoco.json'
    train_dataset = kwcoco.CocoDataset(coco_fpath)
    chan_info = kwcoco_extensions.coco_channel_stats(train_dataset)

    single_channels_hist = ub.ddict(lambda: 0)
    for chans, freq in chan_info['chan_hist'].items():
        for c in kwcoco.FusedChannelSpec.coerce(chans).as_list():
            single_channels_hist[c] += freq

    print('single_channels_hist = {}'.format(ub.urepr(single_channels_hist, nl=1)))

    channel_groups = [
        'red|green|blue',
        'built_up|forest|water',
        'matseg_0|matseg_1|matseg_2',
        'invariants.0:3',
        'panchromatic',
        'depth',
        'nir|swir16|swir22',
        'lwir11|pan|lwir12'
    ]

    channels = ','.join(channel_groups)
    batch_size = 1
    time_steps = 9
    chip_size = 416
    datamodule = KWCocoVideoDataModule(
        train_dataset=train_dataset,
        test_dataset=None,
        batch_size=batch_size,
        channels=channels,
        num_workers=0,
        time_steps=time_steps,
        chip_size=chip_size,
        neg_to_pos_ratio=0,
        normalize_inputs=1,
        use_centered_positives=True,
        use_grid_positives=False,
        resample_invalid_frames=False,
    )
    datamodule.setup('fit')
    dataset = datamodule.torch_datasets['train']

    coco_dset = train_dataset

    # tr = find_varied_region(coco_dset)

    space_slice = (slice(45, 462, None), slice(850, 1267, None))
    gids = [811, 803, 763, 749, 783, 787, 778]

    vidid = coco_dset.index.name_to_video['NZ_R001']['id']
    tr = {
        'main_idx': 0,
        'video_id': vidid,
        # 'space_slice': (slice(0, 96, None), slice(0, 96, None)),
        'space_slice': space_slice,
        'gids': gids,
    }

    dataset.disable_augmenter = True
    dataset.requested_tasks['change'] = False
    dataset.requested_tasks['saliency'] = False

    item = dataset[tr]

    combinable_extra = [kwcoco.FusedChannelSpec.coerce(g) for g in channel_groups]
    combinable_extra = [p.as_list() for p in combinable_extra if p.numel() == 3]
    canvas = dataset.draw_item(item, draw_weights=False, combinable_extra=combinable_extra, max_dim=416)
    kwplot.imshow(canvas)

    kwimage.imwrite('watch_data.jpg', canvas)

    # fig = kwplot.autoplt().gcf()
    # canvas = kwplot.render_figure_to_image(fig)


def assert_temporal_sampler_consistency(dataset):
    coco_dset = dataset.sampler.dset
    vidid_to_imgs = ub.group_items(coco_dset.dataset['images'], key=lambda img: img['video_id'])
    for vidid, temporal_sampler in dataset.new_sample_grid['vidid_to_time_sampler'].items():
        # vidids = coco_dset.images(temporal_sampler.video_gids).lookup('video_id')
        imgs = vidid_to_imgs[vidid]
        n_imgs = len(imgs)
        n_video_gids = len(temporal_sampler.video_gids)
        assert n_video_gids == n_imgs
        print(f'{vidid=} {n_imgs=} {n_video_gids=}')


def find_varied_region(coco_dset, dataset):
    vidid = coco_dset.index.name_to_video['NZ_R001']['id']
    # vidid = list(coco_dset.videos())[3]
    coco_dset.index.videos[vidid]
    images = coco_dset.images(vidid=vidid)
    sensor_gids = ub.ddict(list)
    for coco_img in images.coco_images:
        sensor = coco_img.img.get('sensor_coarse')
        sensor_gids[sensor].append(coco_img.img['id'])

    import kwarray
    for sensor in sensor_gids.keys():
        sensor_gids[sensor] = kwarray.shuffle(sensor_gids[sensor])

    for gid in sensor_gids['WV']:
        if 'depth' in coco_dset.coco_image(gid).channels:
            print(gid)
            pass
        pass

    gids = list(ub.flatten(list(zip(*sensor_gids.values()))))[0:9]
    coco_dset.images(gids).lookup('sensor_coarse')

    pos_idxs = dataset.new_sample_grid['positives_indexes']
    targets = dataset.new_sample_grid['targets']
    space_slices = []
    for idx in pos_idxs:
        tr = targets[idx]
        if tr['video_id'] == vidid:
            print('tr = {!r}'.format(tr))
            space_slices.append(tr['space_slice'])

    slice_options = list(ub.unique(space_slices, key=ub.hash_data))
    print('slice_options = {}'.format(ub.urepr(slice_options, nl=1)))
    space_slice = slice_options[0]

    # gids = [843, 923, 921, 927]

    # for gid in sensor_gids['WV']:
    #     coco_img = coco_dset.coco_image(gid)
    #     data = coco_img.imdelay(channels='panchromatic').finalize(space='video')

    # for gid in sensor_gids['WV']:
    #     print(coco_dset.imgs[gid]['valid_region'])

    # gids = list(sensor_gids.values())
    # print('gids = {!r}'.format(gids))
    space_slice = (slice(45, 462, None), slice(850, 1267, None))
    gids = [811, 803, 763, 823, 749, 783, 796, 787, 778]

    tr = {
        'main_idx': 0,
        'video_id': vidid,
        # 'space_slice': (slice(0, 96, None), slice(0, 96, None)),
        'space_slice': space_slice,
        'gids': gids,
    }
    return tr



import argparse
import sys

import ubelt
import kwcoco


def main():
    parser = argparse.ArgumentParser(
        description="Concatenate KWCOCO datasets at the video level")

    parser.add_argument("--dst",
                        type=str,
                        help="Path to output KWCOCO dataset")
    parser.add_argument('src_kwcoco_datasets',
                        type=str,
                        nargs='+',
                        help="Paths to input KWCOCO datasets to concatenate")

    concat_kwcoco_datasets(**vars(parser.parse_args()))


def concat_kwcoco_datasets(src_kwcoco_datasets, dst):
    first, *rest = src_kwcoco_datasets

    out_dset = kwcoco.CocoDataset(first).copy()

    def _build_video_lookup(kwcoco_dset):
        video_metas = kwcoco_dset.videos().lookup(('id', 'name'))
        return dict(zip(video_metas['name'], video_metas['id']))

    out_dset_videos_lookup = _build_video_lookup(out_dset)

    for dset_path in rest:
        next_id = max(ubelt.flatten(out_dset.videos().images.lookup('id')))
        dset = kwcoco.CocoDataset(dset_path)

        dset_videos_lookup = _build_video_lookup(dset)

        for video_name, video_id in dset_videos_lookup.items():
            if video_name in out_dset_videos_lookup:
                out_video_id = out_dset_videos_lookup[video_name]

                for image_group in dset.videos([video_id]).images:
                    for image in image_group.objs:
                        next_id += 1
                        image['id'] = next_id
                        image['video_id'] = out_video_id

                        out_dset.add_image(**image)
            else:
                out_dset.add_video(dset.videos([video_id]))

    out_dset.dump(dst, indent=2, newlines=True)


if __name__ == "__main__":
    sys.exit(main())



# Code for different variants of U-Net
# Some parts taken from https://github.com/milesial/Pytorch-UNet
# Implements light (half feature channels) and lighter (quarter number of
# feature maps) U-Net


import torch
import torch.nn as nn
import torch.nn.functional as F


def count_trainable_parameters(model):  # to count trainable parameters
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


class double_conv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(double_conv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class inconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(inconv, self).__init__()
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x):
        x = self.conv(x)
        return x


class down(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(down, self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool2d(2),
            double_conv(in_ch, out_ch)
        )

    def forward(self, x):
        x = self.mpconv(x)
        return x


class up(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(up, self).__init__()
        self.up = nn.Upsample(
            scale_factor=2,
            mode='bilinear',
            align_corners=True)

        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x1, x2):
        #x1 = self.up(x1)
        x1 = nn.functional.interpolate(x1, scale_factor=2, mode='nearest')

        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(
            x1,
            (diffX // 2,
             diffX - diffX // 2,
             diffY // 2,
             diffY - diffY // 2))

        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x


class outconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(outconv, self).__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, 1, bias=False)

    def forward(self, x):
        x = self.conv(x)
        return x


# U-Net
class UNet(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(UNet, self).__init__()

        self.inc = inconv(in_channels, 64)
        self.down1 = down(64, 128)
        self.down2 = down(128, 256)
        self.down3 = down(256, 512)
        self.down4 = down(512, 512)
        self.up1 = up(1024, 256)
        self.up2 = up(512, 128)
        self.up3 = up(256, 64)
        self.up4 = up(128, 64)
        self.outc = outconv(64, out_channels)

    def forward(self, x):
        # x1 = self.inc(x.permute(0, 3, 1, 2))
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)

        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.outc(x)
        return x


class UNetEncoder(nn.Module):
    def __init__(self, in_channels):
        super(UNetEncoder, self).__init__()

        self.inc = inconv(in_channels, 64)
        self.down1 = down(64, 128)
        self.down2 = down(128, 256)
        self.down3 = down(256, 512)
        self.down4 = down(512, 512)

    def forward(self, x):
        # x1 = self.inc(x.permute(0, 3, 1, 2))
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        return x1, x2, x3, x4, x5


class UNetDecoder(nn.Module):
    def __init__(self, out_channels):
        super(UNetDecoder, self).__init__()

        self.up1 = up(1024, 256)
        self.up2 = up(512, 128)
        self.up3 = up(256, 64)
        self.up4 = up(128, 64)
        self.outc = outconv(64, out_channels)

    def forward(self, X):
        x1, x2, x3, x4, x5 = X
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.outc(x)
        return x


class UNetHalfsizeEncoder(nn.Module):
    def __init__(self, in_channels):
        super(UNetEncoder, self).__init__()

        self.inc = inconv(in_channels, 32)
        self.down1 = down(32, 64)  # down(64, 128)
        self.down2 = down(64, 128)  # down(128, 256)
        self.down3 = down(128, 256)  # down(256, 512)
        self.down4 = down(256, 256)  # down(512, 512)

    def forward(self, x):
        # x1 = self.inc(x.permute(0, 3, 1, 2))
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        return x1, x2, x3, x4, x5


class UNetHalfsizeDecoder(nn.Module):
    def __init__(self, out_channels):
        super(UNetDecoder, self).__init__()

        self.up1 = up(512, 128)  # up(1024, 256)
        self.up2 = up(256, 64)  # up(512, 128)
        self.up3 = up(128, 32)  # up(256, 64)
        self.up4 = up(64, 32)  # up(128, 64)
        self.outc = outconv(32, out_channels)

    def forward(self, X):
        x1, x2, x3, x4, x5 = X
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.outc(x)
        return x
